# Copyright 2021 Edward Leardi. All Rights Reserved.
#
# Copyright 2020 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"High-level functions tailored for new users to quickly get started with Nourish for common tasks."


# We don't use __all__ in this file because having every exposed function shown in __init__.py is more clear.

from copy import deepcopy
import dataclasses
import functools
from textwrap import dedent
from typing import Any, Callable, Dict, Iterable, Optional, Tuple, TypeVar, Union, cast
from packaging.version import parse as version_parser

from ._config import Config
from ._dataset import Dataset
from . import typing as typing_
from ._schema import Schema, SchemaDict, SchemaManager

# Global configurations --------------------------------------------------

_global_config: Config


def get_config() -> Config:
    """Returns global Nourish configs.

    :return: Read-only global configs represented as a data class

    Example:

    >>> get_config()
    Config(DATASET_SCHEMA_URL=..., FORMAT_SCHEMA_URL=..., LICENSE_SCHEMA_URL=..., DATADIR=...)
    """
    return _global_config


# The SchemaManager object that is managed by high-level functions
_schemata: Optional[SchemaManager] = None


def init(update_only: bool = True, **kwargs: Any) -> None:
    """
    (Re-)initialize high-level functions and their configurations.

    :param update_only: If ``True``, only update in the global configs what config is specified; reuse schemata loaded
        by high-level functions if URLs do not change. Otherwise, reset everything to default in global configs except
        those specified as keyword arguments; clear all schemata loaded by high-level functions.
    :param DATASET_SCHEMA_URL: The default dataset schema file URL.
    :param FORMAT_SCHEMA_URL: The default format schema file URL.
    :param LICENSE_SCHEMA_URL: The default license schema file URL.
    :param DATADIR: Default dataset directory to download/load to/from. The path can be either absolute or relative to
        the current working directory, but will be converted to the absolute path immediately in this function.
        Defaults to: :file:`~/.nourish/data`.
    """
    global _global_config, _schemata

    if update_only:
        # We don't use dataclasses.replace here because it is uncertain whether it would work well with
        # pydantic.dataclasses.
        prev = dataclasses.asdict(_global_config)
        prev.update(kwargs)
        _global_config = Config(**prev)
    else:
        _global_config = Config(**kwargs)
        _schemata = None


init(update_only=False)

# Dataset --------------------------------------------------


def list_all_datasets() -> Dict[str, Tuple]:
    """Show all available nourish datasets and their versions.

    :return: Mapping of available datasets and their versions.

    Example:

    >>> import pprint
    >>> datasets = list_all_datasets()
    >>> pprint.pprint(datasets)
    {...'gmb': ('1.0.2',),... 'wikitext103': ('1.0.1',)...}
    """

    dataset_schema = export_schemata().schemata['datasets'].export_schema('datasets')
    return {
        outer_k: tuple(inner_k for inner_k, inner_v in outer_v.items())
        for outer_k, outer_v in dataset_schema.items()
    }


# We would like to be more specific about what this function does and avoid using "cast", but this seems to be the best
# we can do at this moment: It at least preserves all type hints of the decorated functions. When callback protocols
# come out, we may be able to make improvements over this part:
# https://mypy.readthedocs.io/en/stable/protocols.html#callback-protocols
_DecoratedFuncType = TypeVar("_DecoratedFuncType", bound=Callable)


def _handle_name_param(func: _DecoratedFuncType) -> _DecoratedFuncType:
    """Decorator for handling ``name`` parameter.

    :raises TypeError: ``name`` is not a string.
    :raises KeyError: ``name`` is not a valid Nourish dataset name.
    :return: Wrapped function that handles ``name`` parameter properly.
    """
    @functools.wraps(func)
    def name_wrapper(name: str, *args: Any, **kwargs: Any) -> Any:

        if not isinstance(name, str):
            raise TypeError('The name parameter must be supplied a str.')
        all_datasets = list_all_datasets()
        if name not in all_datasets.keys():
            raise KeyError(f'"{name}" is not a valid Nourish dataset. You can view all valid datasets and their '
                           'versions by running the function nourish.list_all_datasets().')
        return func(name, *args, **kwargs)
    return cast(_DecoratedFuncType, name_wrapper)


def _handle_version_param(func: _DecoratedFuncType) -> _DecoratedFuncType:
    """Decorator for handling ``version`` parameter. Must still be supplied a dataset ``name``.

    :raises TypeError: ``version`` is not a string.
    :raises KeyError: ``version`` is not a valid Nourish version of ``name``.
    :return: Wrapped function that handles ``version`` parameter properly.
    """
    @functools.wraps(func)
    def version_wrapper(name: str, version: str = 'latest', *args: Any, **kwargs: Any) -> Any:

        if not isinstance(version, str):
            raise TypeError('The version parameter must be supplied a str.')
        all_datasets = list_all_datasets()
        if version == 'latest':
            # Grab latest available version
            version = str(max(version_parser(v) for v in all_datasets[name]))
        elif version not in all_datasets[name]:
            raise KeyError(f'"{version}" is not a valid Nourish version for the dataset "{name}". You can view all '
                           'valid datasets and their versions by running the function nourish.list_all_datasets().')
        return func(name=name, version=version, *args, **kwargs)
    return cast(_DecoratedFuncType, version_wrapper)


@_handle_name_param
@_handle_version_param
def load_dataset(name: str, *,
                 version: str = 'latest',
                 download: bool = True,
                 subdatasets: Union[Iterable[str], None] = None) -> Dict[str, Any]:
    """High level function that wraps :class:`dataset.Dataset` class's load and download functionality. Downloads to and
    loads from directory: :file:`DATADIR/schema_name/name/version` where ``DATADIR`` is in
    ``nourish.get_config().DATADIR``. ``DATADIR`` can be changed by calling :func:`init`.

    :param name: Name of the dataset you want to load from Nourish's available datasets. You can get a list of these
        datasets by calling :func:`list_all_datasets`.
    :param version: Version of the dataset to load. Latest version is used by default. You can get a list of all
        available versions for a dataset by calling :func:`list_all_datasets`.
    :param download: Whether or not the dataset should be downloaded before loading.
    :param subdatasets: An iterable containing the subdatasets to load. ``None`` means all subdatasets.
    :raises FileNotFoundError: The dataset files were not previously downloaded or can't be found, and ``download`` is
        ``False``.
    :return: Dictionary that holds all subdatasets.

    Example:

    >>> data = load_dataset('noaa_jfk')
    >>> data['jfk_weather_cleaned'][['DATE', 'HOURLYVISIBILITY', 'HOURLYDRYBULBTEMPF']].head(3)
                     DATE  HOURLYVISIBILITY  HOURLYDRYBULBTEMPF
    0 2010-01-01 01:00:00               6.0                33.0
    1 2010-01-01 02:00:00               6.0                33.0
    2 2010-01-01 03:00:00               5.0                33.0
    """

    schemata = export_schemata().schemata['datasets']
    schema = schemata.export_schema('datasets', name, version)
    dataset_schema_name = schemata.export_schema().get('name', 'default')

    data_dir = get_config().DATADIR / dataset_schema_name / name / version
    dataset = Dataset(schema=schema, data_dir=data_dir, mode=Dataset.InitializationMode.LAZY)
    if download and not dataset.is_downloaded():
        dataset.download()
    try:
        return dataset.load(subdatasets=subdatasets)
    except RuntimeError as e:
        raise RuntimeError('Failed to load the dataset because some files are not found. '
                           'Did you forget to download the dataset (by specifying `download=True`)?'
                           f'\nCaused by:\n{e}')


@_handle_name_param
@_handle_version_param
def get_dataset_metadata(name: str, *, version: str = 'latest') -> SchemaDict:
    """Return a dataset's metadata either in human-readable form or as a copy of its schema.

    :param name: Name of the dataset you want get the metadata of. You can get a list of these
        datasets by calling :func:`list_all_datasets`.
    :param version: Version of the dataset to load. Latest version is used by default. You can get a list of all
        available versions for a dataset by calling :func:`list_all_datasets`.
    :return: A dataset's metadata.

    Example:

    >>> import pprint
    >>> metadata = get_dataset_metadata('gmb')
    >>> metadata['name']
    'Groningen Meaning Bank Modified'
    >>> metadata['description']
    'A dataset of multi-sentence texts, together with annotations for parts-of-speech...
    >>> pprint.pprint(metadata['subdatasets'])
    {'gmb_subset_full': {'description': 'A full version of the raw dataset. Used '
                                        'to train MAX model – Named Entity Tagger.',
                         'format': 'txt',
                         'name': 'GMB Subset Full',
                         'path': 'groningen_meaning_bank_modified/gmb_subset_full.txt'}}
    """

    return export_schemata().schemata['datasets'].export_schema('datasets', name, version)


@_handle_name_param
@_handle_version_param
def describe_dataset(name: str, *, version: str = 'latest') -> str:
    """Describe a dataset's metadata in human language. Parameters mean the same as :func:`.get_dataset_metadata`.

    :param name: Name of the dataset you want get the metadata of. You can get a list of these datasets by calling
        :func:`list_all_datasets`.
    :param version: Version of the dataset to load. Latest version is used by default. You can get a list of all
        available versions for a dataset by calling :func:`list_all_datasets`.
    :return: The description.

    Example:

    >>> print(describe_dataset('gmb'))
    Dataset name: Groningen Meaning Bank Modified
    Description: A dataset of multi-sentence texts, ...
    Size: 10M
    Published date: 2019-12-19
    License: Community Data License Agreement – Sharing, Version 1.0 (CDLA-Sharing-1.0)
    Available subdatasets: gmb_subset_full
    """

    schema_manager = export_schemata()
    dataset_schema = schema_manager.schemata['datasets'].export_schema('datasets', name, version)
    license_schema = schema_manager.schemata['licenses'].export_schema('licenses')
    return dedent(f'''
            Dataset name: {dataset_schema["name"]}
            Description: {dataset_schema["description"]}
            Size: {dataset_schema["estimated_size"]}
            Published date: {dataset_schema["published"]}
            License: {license_schema[dataset_schema["license"]]["name"]}
            Available subdatasets: {", ".join(dataset_schema["subdatasets"].keys())}
    ''').strip()


# Schemata --------------------------------------------------


def export_schemata() -> SchemaManager:
    """Return a copy of the :class:`schema.SchemaManager` object managed by high-level functions.

    :return: A copy of the :class:`schema.SchemaManager` object

    Example:

    >>> schema_manager = export_schemata()
    >>> schema_manager.schemata
    {'datasets': ..., 'formats': ..., 'licenses':...}
    """

    return deepcopy(_get_schemata())


def load_schemata(*, force_reload: bool = False, tls_verification: Union[bool, typing_.PathLike] = True) -> None:
    """Loads a :class:`schema.SchemaManager` object that stores all schemata. To export the loaded
    :class:`schema.SchemaManager` object, please use :func:`export_schemata`.

    :param force_reload: If ``True``, force reloading even if the provided URLs by :func:`init` are the same as
         provided last time. Otherwise, only those that are different from previous given ones are reloaded.
    :param tls_verification: Same as ``tls_verification`` in :class:`schema.Schema`.
    :raises ValueError: See :class:`schema.Schema`.
    :raises InsecureConnectionError: See :class:`schema.Schema`.

    Example:
    >>> load_schemata()
    >>> loaded_schemata = export_schemata()
    >>> loaded_schemata.schemata
    {'datasets': ..., 'formats': ..., 'licenses':...}
    """

    urls = {
        'datasets': get_config().DATASET_SCHEMA_URL,
        'formats': get_config().FORMAT_SCHEMA_URL,
        'licenses': get_config().LICENSE_SCHEMA_URL
    }

    global _schemata
    if force_reload or _schemata is None:  # Force reload or clean slate, create a new SchemaManager object
        _schemata = SchemaManager(**{
            name: Schema(url, tls_verification=tls_verification) for name, url in urls.items()})
    else:
        for name, schema in _schemata.schemata.items():
            if schema.retrieved_url_or_path != urls[name]:
                _schemata.add_schema(name, Schema(urls[name], tls_verification=tls_verification))


def _get_schemata() -> SchemaManager:
    """Return the :class:`SchemaManager` object managed by high-level functions. If it is not created, create it. This
    function is used by high-level APIs but it should not be a high-level function itself. It should only be used
    internally when the need to modify the managed :class`SchemaManager` object arises. It should not be exposed for the
    same reason: users should not have this easy access to modify the managed :class`SchemaManager` object."""

    global _schemata

    load_schemata()

    # The return value is guranteed to be SchemaManager instead of Optional[SchemaManager] after load_schemata
    assert _schemata is not None  # nosec: We use assertion for code clarity and mypy detection of _schemata's type

    return _schemata



# Copyright 2021 Edward Leardi. All Rights Reserved.
#
# Copyright 2020 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import dataclasses
import json
import pathlib
import re

from packaging.version import parse as version_parser
import pytest
from pydantic import ValidationError

from nourish import (describe_dataset, export_schemata, get_config, get_dataset_metadata, init, list_all_datasets,
                     load_dataset, load_schemata)
from nourish.dataset import Dataset
from nourish._config import Config
from nourish._high_level import _get_schemata

# Global configurations --------------------------------------------------


class TestGetConfig:
    "Test high-level get_config function."

    def test_default_data_dir(self):
        "Test default data dir."

        nourish_data_home = pathlib.Path.home() / '.nourish' / 'data'
        assert get_config().DATADIR == nourish_data_home
        assert isinstance(get_config().DATADIR, pathlib.Path)


class TestInit:
    "Test high-level init function."

    def test_custom_data_dir(self, tmp_path, wikitext103_schema):
        "Test to make sure Dataset constructor uses new global data dir if one was supplied earlier to nourish.init."

        init(DATADIR=tmp_path)
        assert get_config().DATADIR == tmp_path
        assert isinstance(get_config().DATADIR, pathlib.Path)
        wikitext = Dataset(wikitext103_schema, data_dir=tmp_path, mode=Dataset.InitializationMode.LAZY)
        assert wikitext._data_dir == tmp_path
        assert isinstance(wikitext._data_dir, pathlib.Path)

    def test_custom_relative_data_dir(self, chdir_tmp_path, tmp_sub_dir, tmp_relative_sub_dir):
        "Test using a custom relative data directory."

        init(DATADIR=tmp_relative_sub_dir)
        assert get_config().DATADIR == tmp_sub_dir
        assert get_config().DATADIR.is_absolute()

    def test_custom_symlink_data_dir(self, tmp_symlink_dir):
        "Test using a custom symlink data directory. The symlink should not be resolved."

        init(DATADIR=tmp_symlink_dir)
        assert get_config().DATADIR == tmp_symlink_dir

    def test_non_path_data_dir(self):
        "Test exception when a nonpath is passed as DATADIR."

        with pytest.raises(ValidationError) as e:
            init(DATADIR=10)

        assert re.search((r'1 validation error for Config\s+DATADIR\s+value'
                          r' is not a valid path \(type=type_error.path\)'), str(e.value))

    def test_custom_configs(self):
        "Test custom configs."

        init(update_only=False)  # set back everything to default
        assert dataclasses.asdict(get_config()) == dataclasses.asdict(Config())

        new_urls = {
            'DATASET_SCHEMA_URL': 'some/local/file',
            'FORMAT_SCHEMA_URL': 'file://c:/some/other/local/file',
            'LICENSE_SCHEMA_URL': 'http://some/remote/file'
        }
        init(update_only=True, **new_urls)

        for url, val in new_urls.items():
            assert getattr(get_config(), url) == val
        assert get_config().DATADIR == Config.DATADIR

# Dataset --------------------------------------------------


class TestListAllDatasets:
    "Test high-level list_all_datasets function."

    def test_list_all_datasets(self):
        "Test to make sure test_list_all_datasets function returns available dataset names."

        datasets = list_all_datasets()
        assert frozenset(datasets.keys()) == frozenset(['gmb', 'noaa_jfk', 'wikitext103'])


class TestLoadDataset:
    "Test high-level load_dataset function."

    def test_name_param(self, tmp_path):
        "Test to see the name parameter is being handled properly."

        init(DATADIR=tmp_path)

        with pytest.raises(TypeError) as e:
            load_dataset(123)
        assert str(e.value) == 'The name parameter must be supplied a str.'

        name = 'fake_dataset'
        with pytest.raises(KeyError) as e:
            load_dataset(name)
        assert str(e.value) == (f'\'"{name}" is not a valid Nourish dataset. You can view all valid datasets and their '
                                'versions by running the function nourish.list_all_datasets().\'')

    def test_version_param(self, tmp_path):
        "Test to see the version parameter is being handled properly."

        init(DATADIR=tmp_path)

        with pytest.raises(TypeError) as e:
            load_dataset('gmb', version=1.0)
        assert str(e.value) == 'The version parameter must be supplied a str.'

        name, version = 'gmb', ''
        with pytest.raises(KeyError) as e:
            load_dataset('gmb', version=version)
        assert str(e.value) == (f'\'"{version}" is not a valid Nourish version for the dataset "{name}". '
                                'You can view all valid datasets and their versions by running the function '
                                'nourish.list_all_datasets().\'')

        name, version = 'gmb', 'fake_version'
        with pytest.raises(KeyError) as e:
            load_dataset('gmb', version=version)
        assert str(e.value) == (f'\'"{version}" is not a valid Nourish version for the dataset "{name}". '
                                'You can view all valid datasets and their versions by running the function '
                                'nourish.list_all_datasets().\'')

        # If no version specified, make sure latest version grabbed
        all_datasets = list_all_datasets()
        latest_version = str(sorted(version_parser(v) for v in all_datasets[name])[-1])
        assert load_dataset('gmb') == load_dataset('gmb', version=latest_version)

    def test_subdatasets_param(self, tmp_path):
        "Test to see subdatasets parameter is being handled properly."

        init(DATADIR=tmp_path)

        with pytest.raises(TypeError) as e:
            load_dataset('wikitext103', version='1.0.1', download=True, subdatasets=123)
        assert str(e.value) == '\'int\' object is not iterable'

        subdatasets = ['train']
        wikitext103_data = load_dataset('wikitext103', version='1.0.1', download=True, subdatasets=subdatasets)
        assert list(wikitext103_data.keys()) == subdatasets

    def test_default_dataset_schema_name(self, tmp_path, gmb_schema):
        "Test the default schemata name."

        init(DATADIR=tmp_path)
        data_dir = tmp_path / 'default' / 'gmb' / '1.0.2'
        gmb = Dataset(gmb_schema, data_dir=data_dir, mode=Dataset.InitializationMode.DOWNLOAD_AND_LOAD)
        _get_schemata().schemata['datasets']._schema.pop('name')  # Remove the "name" key
        gmb_data = load_dataset('gmb', version='1.0.2', download=False)
        assert gmb.data == gmb_data

    def test_download_false(self, tmp_path, gmb_schema):
        "Test to see the function loads properly when download=False and dataset was previously downloaded."

        init(DATADIR=tmp_path)
        data_dir = tmp_path / 'dax' / 'gmb' / '1.0.2'
        gmb = Dataset(gmb_schema, data_dir=data_dir, mode=Dataset.InitializationMode.DOWNLOAD_AND_LOAD)
        gmb_data = load_dataset('gmb', version='1.0.2', download=False)
        assert gmb.data == gmb_data

    def test_download_true(self, tmp_path, downloaded_gmb_dataset):
        "Test to see the function downloads and loads properly when download=True."

        init(DATADIR=tmp_path)
        downloaded_gmb_dataset_data = downloaded_gmb_dataset.load()
        gmb_data = load_dataset('gmb', version='1.0.2', download=True)
        assert downloaded_gmb_dataset_data == gmb_data

    def test_loading_undownloaded(self, tmp_path):
        "Test loading before ``Dataset.download()`` has been called."

        init(DATADIR=tmp_path)
        with pytest.raises(RuntimeError) as e:
            load_dataset('wikitext103', version='1.0.1', download=False)
        assert 'Did you forget to download the dataset (by specifying `download=True`)?' in str(e.value)


def test_get_dataset_metadata():
    "Test ``get_dataset_metadata``."

    name, version = 'gmb', '1.0.2'

    gmb_schema = get_dataset_metadata(name, version=version)
    assert gmb_schema == export_schemata().schemata['datasets'].export_schema('datasets', name, version)


def test_describe_dataset():
    "Test ``describe_dataset``."

    name, version = 'gmb', '1.0.2'

    gmb_description = describe_dataset(name, version=version)
    dataset_schema = export_schemata().schemata['datasets'].export_schema('datasets', name, version)
    license_schema = export_schemata().schemata['licenses'].export_schema('licenses')

    # Check a couple of spots
    assert dataset_schema['name'] in gmb_description
    assert dataset_schema['estimated_size'] in gmb_description
    assert license_schema[dataset_schema["license"]]["name"] in gmb_description

    # Instead of copying over the string for testing, we test a couple of important characteristics here
    gmb_lines = gmb_description.splitlines()
    assert len(gmb_lines) == 6  # number of lines
    assert all(line.strip() != '' for line in gmb_lines)  # no blank line
    assert all(line.strip() == line for line in gmb_lines)  # no trailing or leading whitespace
    assert all(':' in line for line in gmb_lines)  # no missing colons

# Schemata --------------------------------------------------


class TestSchemataFunctions:
    "Test the high-level schemata functions."

    def test_load_schemata(self, loaded_schemata, schema_file_absolute_dir):
        "Test high-level load_schemata function."

        init(update_only=False,
             DATASET_SCHEMA_URL=loaded_schemata.schemata['datasets'].retrieved_url_or_path,
             FORMAT_SCHEMA_URL=loaded_schemata.schemata['formats'].retrieved_url_or_path,
             LICENSE_SCHEMA_URL=loaded_schemata.schemata['licenses'].retrieved_url_or_path)
        load_schemata(force_reload=True)
        for name in ('datasets', 'formats', 'licenses'):
            assert (_get_schemata().schemata[name].retrieved_url_or_path ==
                    loaded_schemata.schemata[name].retrieved_url_or_path)

        init(update_only=True,
             # Different from the previous relative path used in loaded_schemata
             DATASET_SCHEMA_URL=schema_file_absolute_dir / 'datasets.yaml')
        load_schemata(force_reload=False)
        for name in ('formats', 'licenses'):
            assert (_get_schemata().schemata[name].retrieved_url_or_path ==
                    loaded_schemata.schemata[name].retrieved_url_or_path)
        assert _get_schemata().schemata['datasets'].retrieved_url_or_path == schema_file_absolute_dir / 'datasets.yaml'

    def test_export_schemata(self, schema_file_absolute_dir, schema_file_https_url):
        "Test high-level export-schemata function."

        assert export_schemata() is not _get_schemata()
        # The two returned schemata should equal
        assert (json.dumps(export_schemata().schemata['datasets'].export_schema(),
                           sort_keys=True, indent=2, default=str) ==
                json.dumps(_get_schemata().schemata['datasets'].export_schema(), sort_keys=True, indent=2, default=str))

        # Different from https url used by nourish_initialization autouse fixture
        new_urls = {
            'DATASET_SCHEMA_URL': schema_file_absolute_dir / 'datasets.yaml',
            'LICENSE_SCHEMA_URL': schema_file_absolute_dir / 'licenses.yaml'
        }
        init(update_only=True, **new_urls)
        assert export_schemata().schemata['formats'].retrieved_url_or_path == f'{schema_file_https_url}/formats.yaml'
        assert export_schemata().schemata['datasets'].retrieved_url_or_path == new_urls['DATASET_SCHEMA_URL']
        assert export_schemata().schemata['licenses'].retrieved_url_or_path == new_urls['LICENSE_SCHEMA_URL']



# Copyright 2021 Edward Leardi. All Rights Reserved.
#
# Copyright 2020 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import copy
import hashlib
from http.server import HTTPServer, SimpleHTTPRequestHandler
import os
from pathlib import Path
from ssl import PROTOCOL_TLS_SERVER, SSLContext
from tempfile import TemporaryDirectory
import threading
from typing import Callable
from urllib.request import urlretrieve
import uuid

import certifi
import pytest

from nourish import init
from nourish._high_level import _get_schemata
from nourish.dataset import Dataset
from nourish.schema import Schema, SchemaDict, SchemaManager

# Basic utilities --------------------------------


@pytest.fixture
def tmp_sub_dir(tmp_path):
    "A ``pathlib.Path`` object that points to a temporary dir, created as a subdir of ``tmp_path``."

    with TemporaryDirectory(dir=tmp_path) as d:
        yield Path(d).absolute()


@pytest.fixture
def tmp_relative_sub_dir(tmp_path, tmp_sub_dir):
    """The path of ``tmp_sub_dir`` relative to ``tmp_dir``. This can be useful (in conjunction with ``chdir_tmp_path``)
    for some tests that focus on an absolute path vis-a-vis its relative path form. In these tests, the use of
    ``os.path.relpath()`` is discouraged, because it would fail if the input is on a different drive from the working
    directory on Windows. Instead, use `tmp_sub_dir` and ``tmp_relative_sub_dir``."""

    return tmp_sub_dir.relative_to(tmp_path)


@pytest.fixture
def tmp_symlink_dir(tmp_path, tmp_sub_dir):
    "A ``pathlib.Path`` object that points to a temporary symlink to ``tmp_sub_dir``. It always sits in ``tmp_dir``."

    while True:
        symlink_dir = tmp_path / str(uuid.uuid4())
        # A collision is extremely unlikely, but for the sake of completeness, this check never hurts.
        if not symlink_dir.exists():
            break

    symlink_dir.symlink_to(tmp_sub_dir, target_is_directory=True)

    yield symlink_dir

    symlink_dir.unlink()


@pytest.fixture
def chdir_tmp_path(tmp_path):
    "Make the test run in ``tmp_path`` as the current working directory."

    cur_dir = Path.cwd()
    os.chdir(tmp_path)
    yield
    os.chdir(cur_dir)


@pytest.fixture(scope='session')
def local_http_server() -> HTTPServer:
    "A local http server that serves the source directory."

    with HTTPServer(("localhost", 8080), SimpleHTTPRequestHandler) as httpd:
        # Start a new thread, because httpd.serve_forever is blocking
        threading.Thread(target=httpd.serve_forever, name='Local Http Server', daemon=True).start()
        yield httpd
        httpd.shutdown()


@pytest.fixture(scope='session')
def local_http_server_root_url(local_http_server) -> str:
    "Root URL of the local http server."

    return f'http://{local_http_server.server_address[0]}:{local_http_server.server_address[1]}'


@pytest.fixture(scope='session')
def local_https_server() -> HTTPServer:
    "A local https server that serves the source directory."

    # Merge certifi's CA bundle (used as default by requests) with ours. This is done by a simple concatenation of the
    # two pem files. Although pem files are text files, we treat them as binaries to avoid unexpected EOL conversions.
    ca_bundle = Path('tests/tls/test_ca_bundle.pem')
    ca_bundle.write_bytes(
        Path(certifi.where()).read_bytes() + Path('tests/tls/server.pem').read_bytes())
    os.environ['REQUESTS_CA_BUNDLE'] = str(ca_bundle)
    # Ensure CURL_CA_BUNDLE isn't involved
    os.environ.pop('CURL_CA_BUNDLE', None)

    with HTTPServer(("localhost", 8081), SimpleHTTPRequestHandler) as httpd:
        context = SSLContext(PROTOCOL_TLS_SERVER)
        context.load_cert_chain('tests/tls/server.pem', keyfile='tests/tls/server.key')
        httpd.socket = context.wrap_socket(httpd.socket, server_side=True)
        # Start a new thread, because httpd.serve_forever is blocking
        threading.Thread(target=httpd.serve_forever, name='Local Https Server', daemon=True).start()
        yield httpd
        httpd.shutdown()


@pytest.fixture(scope='session')
def local_https_server_root_url(local_https_server) -> str:
    "Root URL of the local https server."

    return f'https://{local_https_server.server_address[0]}:{local_https_server.server_address[1]}'


@pytest.fixture
def untrust_self_signed_cert(local_https_server):
    # local_https_server is here to ensure that environment variables have been initialized
    """Untrust our self signed TLS certificate of our test local HTTPS server, and let requests uses the default trusted
    CA bundle."""
    self_signed = os.environ.pop('REQUESTS_CA_BUNDLE')
    yield
    os.environ['REQUESTS_CA_BUNDLE'] = self_signed


@pytest.fixture(autouse=True)
def nourish_initialization(schema_file_https_url, schema_localized_url):
    """Create the default initialization used for all tests. This is mainly for having a uniform initialization for all
    tests as well as avoiding using the actual default schema file URLs so as to decouple the two lines of development
    (default schema files and this library). It also replaces all download URLs with localized URLs."""

    init(update_only=False,
         DATASET_SCHEMA_URL=f'{schema_file_https_url}/datasets.yaml',
         FORMAT_SCHEMA_URL=f'{schema_file_https_url}/formats.yaml',
         LICENSE_SCHEMA_URL=f'{schema_file_https_url}/licenses.yaml')

    # Use local dataset locations by default in our tests
    datasets = _get_schemata().schemata['datasets']._schema['datasets']
    for name, versions in datasets.items():
        for version in versions:
            datasets[name][version] = schema_localized_url(name, version)

# Dataset --------------------------------------


@pytest.fixture(scope='session')
def dataset_base_url(local_https_server_root_url) -> str:
    "The base local HTTPS server URL that stores datasets for testing purposes."

    return f'{local_https_server_root_url}/tests/datasets'


@pytest.fixture(scope='session')
def dataset_dir() -> Path:
    "The directory that stores datasets for testing purposes."

    d = Path('tests/datasets')
    d.mkdir(exist_ok=True)

    return d


@pytest.fixture(scope='session')
def _download_dataset(dataset_dir, _loaded_schemata) -> Callable[[str], None]:
    """Utility function for downloading datasets to ``tests/datasets/{name}-{version}`` for testing purpose. These files
    will not be deleted after the test session terminates, and they are cached for future test sessions. Accordingly, if
    ``tests/datasets/{name}-{version}`` is already present, this fixture does nothing.
    """
    # We use _loaded_schemata instead of loaded_schemata to avoid scope mismatch error (a session-scoped fixture can't
    # call a function-scoped fixture)

    def _download_dataset_impl(name, version):
        # we drop the 'tar.gz' extension here -- our package should work regardless of the extension, and we allow the
        # file to be archived in a different compression format.
        local_destination = dataset_dir / f'{name}-{version}.tar.gz'

        schema = _loaded_schemata.schemata['datasets'].export_schema('datasets', name, version)

        if local_destination.exists() and \
           hashlib.sha512(local_destination.read_bytes()).hexdigest() == schema['sha512sum']:
            # The file has been completely downloaded before
            return

        # We use urllib instead of requests to avoid running the same code path with our downloading implementation
        urlretrieve(schema['download_url'], filename=local_destination)

    return _download_dataset_impl

# Schema fixtures --------------------------------------------------


@pytest.fixture(scope='session')
def schema_localized_url(_loaded_schemata, _download_dataset, dataset_base_url) -> Callable[[str, str], SchemaDict]:
    "Utility function fixture for generating schema fixtures with its downloading URL modified to the local HTTPS URL."
    # We use _loaded_schemata instead of loaded_schemata to avoid scope mismatch error (a session-scoped fixture can't
    # call a function-scoped fixture)

    def _schema_localized_url_impl(name, version):
        _download_dataset(name, version)
        schema = _loaded_schemata.schemata['datasets'].export_schema('datasets', name, version)
        schema['download_url'] = str(f'{dataset_base_url}/{name}-{version}.tar.gz')
        return schema

    return _schema_localized_url_impl


@pytest.fixture(scope='session')
def _loaded_schemata(schema_file_relative_dir) -> SchemaManager:
    """A loaded ``SchemaManager`` object, but this should never be modified. This object manages ``Schema`` objects
    corresponding to ``tests/{datasets,formats,licenses}.yaml``. Note that these are not necessarily the same as the
    ones used in other schema fixtures, so please do not assume that it is equal to other schema fixtures. One purpose
    of this fixture is to reduce repeated call in the test to the same function when ``loaded_schemata`` is used. The
    other purpose is to provide other session-scoped fixtures access to the loaded schemata, because session-scoped
    fixtures can't load function-scoped fixtures.
    """

    return SchemaManager(datasets=Schema(schema_file_relative_dir / 'datasets.yaml'),
                         formats=Schema(schema_file_relative_dir / 'formats.yaml'),
                         licenses=Schema(schema_file_relative_dir / 'licenses.yaml'))


@pytest.fixture
def loaded_schemata(_loaded_schemata) -> SchemaManager:
    """A copy of _loaded_schemata. Tests outside this file should always use this one so as to avoid mistakenly
    modifying the content."""

    return copy.deepcopy(_loaded_schemata)


# Every _*_schema fixture also implies that a session wide test dataset file is downloaded. They should only be read
# because we want them to be session-scoped. All tests should use the fixture without a leading underscore.

# We don't create a function that automatically generates all the following fixtures because explicitly listing them
# would be easier to understand the error when a test fails.

@pytest.fixture(scope='session')
def _gmb_schema(schema_localized_url):
    return schema_localized_url('gmb', '1.0.2')


@pytest.fixture
def gmb_schema(_gmb_schema):
    return copy.deepcopy(_gmb_schema)


@pytest.fixture(scope='session')
def _noaa_jfk_schema(schema_localized_url):
    return schema_localized_url('noaa_jfk', '1.1.4')


@pytest.fixture
def noaa_jfk_schema(_noaa_jfk_schema):
    return copy.deepcopy(_noaa_jfk_schema)


@pytest.fixture(scope='session')
def _wikitext103_schema(schema_localized_url):
    return schema_localized_url('wikitext103', '1.0.1')


@pytest.fixture
def wikitext103_schema(_wikitext103_schema):
    return copy.deepcopy(_wikitext103_schema)


# Schema file locations fixture --------------------------------------


@pytest.fixture(scope='session')
def schema_file_absolute_dir() -> Path:
    "The base of the absolute path to the dir that contains test schema files."

    return Path.cwd() / 'tests' / 'schemata'


@pytest.fixture(scope='session')
def schema_file_relative_dir() -> Path:
    "The base of the relative path to the dir that contains test schema files."

    return Path('tests/schemata')


@pytest.fixture(scope='session')
def schema_file_file_url(schema_file_absolute_dir) -> str:
    "The base of file:// schema file URLs."

    return schema_file_absolute_dir.as_uri()


@pytest.fixture(scope='session')
def schema_file_http_url(local_http_server_root_url) -> str:
    "The base of remote http:// test schema file URLs."

    return f"{local_http_server_root_url}/tests/schemata"


@pytest.fixture(scope='session')
def schema_file_https_url(local_https_server_root_url) -> str:
    "The base of remote https:// test schema file URLs."

    return f"{local_https_server_root_url}/tests/schemata"


# Downloaded datasets -------------------------------------


@pytest.fixture
def downloaded_wikitext103_dataset(wikitext103_schema) -> Dataset:
    with TemporaryDirectory() as tmp_data_dir:
        yield Dataset(wikitext103_schema, data_dir=tmp_data_dir, mode=Dataset.InitializationMode.DOWNLOAD_ONLY)


@pytest.fixture
def downloaded_gmb_dataset(gmb_schema) -> Dataset:
    with TemporaryDirectory() as tmp_data_dir:
        yield Dataset(gmb_schema, data_dir=tmp_data_dir, mode=Dataset.InitializationMode.DOWNLOAD_ONLY)


@pytest.fixture
def downloaded_noaa_jfk_dataset(noaa_jfk_schema) -> Dataset:
    with TemporaryDirectory() as tmp_data_dir:
        yield Dataset(noaa_jfk_schema, data_dir=tmp_data_dir, mode=Dataset.InitializationMode.DOWNLOAD_ONLY)



# Copyright 2021 Edward Leardi. All Rights Reserved.
#
# Copyright 2020 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"Directory lock."


from contextlib import contextmanager
import itertools
import os
import pathlib
import threading
from uuid import uuid4
from typing import Iterator

from . import typing as typing_


class DirectoryLockAcquisitionError(RuntimeError):
    "Raised when failed to acquire a lock."


class DirectoryLock:
    """Read/write lock for a directory.

    A lock file is named as ``{read|write}.{pid}.{uuid}.lock``. This should be able to resolve all potential name
    clashes. We put process ID here in case someone wants to figure out which process created the file.

    As a reservation for future compatibility, this class reserves all files starting with ``read.``/``write.`` and ends
    with ``.lock`` for its own use.

    :param directory: The directory where lock files would be put.
    """

    def __init__(self, directory: typing_.PathLike):
        self._uuid: str = str(uuid4())
        self._directory: pathlib.Path = pathlib.Path(directory)
        self._thread_lock: threading.Lock = threading.Lock()

    @property
    def _lock_file_suffix(self) -> str:
        "The suffix of the lock file."
        return f'.{os.getpid()}.{self._uuid}.lock'

    def _get_read_locks(self) -> Iterator[pathlib.Path]:
        """Get a list of read lock files.

        :return: An iterable of read lock file paths. Empty if there's no read lock file.
        """
        return self._directory.glob("read.*.lock")

    def _get_write_locks(self) -> Iterator[pathlib.Path]:
        """Get a list of write lock files.

        :return: An iterable of write lock file paths. Empty if there's no write lock file.
        """

        return self._directory.glob("write.*.lock")

    def _does_read_lock_exist(self) -> bool:
        "Returns True if a read lock file exists, otherwise False."
        return next(self._get_read_locks(), None) is not None

    def _does_write_lock_exist(self) -> bool:
        "Returns True if a write lock file exists, otherwise False."
        return next(self._get_write_locks(), None) is not None

    def lock(self, *, write: bool) -> bool:
        """Lock the directory (create the lock file in the directory). If the directory does not exist, create it.

        :param write: Whether this is a write lock or a read lock. A write lock excludes others from both reading and
            writing, while a read lock only excludes writing. Multiple read locks can exist at the same time, but only
            one write lock may exist at any single moment.
        :return: True if lock succeeds, False if fails. This function does not throw exceptions because :meth:`.lock` is
            also used for peeking whether the lock is obtainable.
        """

        lock_file = self._directory / f'{"write" if write else "read"}{self._lock_file_suffix}'

        with self._thread_lock:
            if not self._directory.exists():
                self._directory.mkdir(parents=True)
            if not self._directory.is_dir():
                raise NotADirectoryError(f'"{self._directory}" exists and is not a directory.')
            if write:  # write lock
                if self._does_read_lock_exist() or self._does_write_lock_exist():
                    return False
                else:
                    lock_file.touch(exist_ok=False)
            else:  # read lock
                if self._does_write_lock_exist():
                    return False
                else:
                    lock_file.touch(exist_ok=False)

        return True

    def unlock(self) -> bool:
        """Unlock the directory.

        :return: True if unlock succeeds, False if there is no lock to remove. This function does not throw exceptions
            because it is commonplace to call :meth:`.unlock` from multiple locations and we consider the situation
            where the lock has been removed as an expected usage.
        """
        with self._thread_lock:
            lock_existed: bool = False
            write_lock_file = self._directory / f'write{self._lock_file_suffix}'
            if write_lock_file.exists():
                write_lock_file.unlink()
                lock_existed = True
            read_lock_file = self._directory / f'read{self._lock_file_suffix}'
            if read_lock_file.exists():
                read_lock_file.unlink()
                lock_existed = True
            return lock_existed

    @contextmanager
    def locking(self, *, write: bool) -> Iterator[bool]:
        """Same as :meth:`.lock`, but used in ``with`` statements.

        Example:

        .. code-block:: python

           with some_lock.locking(write=True):
               # do the work ...
        """
        try:
            yield self.lock(write=write)
        finally:
            self.unlock()

    @contextmanager
    def locking_with_exception(self, *, write: bool) -> Iterator[None]:
        """Similar to :meth:`.locking`, but raises exceptions if the lock is failed to acquire.

        :raises DirectoryLockAcquisitionError: Failed to acquire the directory lock.

        Example:

        .. code-block:: python

           with some_lock.locking_with_exception(write=True):
               # do the work ...
        """
        with self.locking(write=write) as lock:
            if not lock:
                raise DirectoryLockAcquisitionError(
                    f'Failed to acquire directory {"write" if write else "read"} lock for "{self._directory}"')
            yield

    def force_clear_all_locks(self) -> None:
        """Force clear all read locks. This is useful in situations such as those when system crashes and locks must be
        reset.
        """

        for f in itertools.chain(self._get_read_locks(), self._get_write_locks()):
            os.remove(f)



# Copyright 2021 Edward Leardi. All Rights Reserved.
#
# Copyright 2020 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from urllib.parse import urlparse

import pytest
import requests.exceptions

from nourish import init
from nourish._config import Config
from nourish._schema_retrieval import retrieve_schema_file


class TestConfig:
    "Test the Config dataclass."

    def test_default_schema_url_https(self):
        "Test the default schema URLs are https-schemed."

        assert urlparse(Config.DATASET_SCHEMA_URL).scheme == 'https'
        assert urlparse(Config.FORMAT_SCHEMA_URL).scheme == 'https'
        assert urlparse(Config.LICENSE_SCHEMA_URL).scheme == 'https'

    @pytest.mark.xfail(reason="default remote might be down but it's not this library's issue",
                       raises=requests.exceptions.ConnectionError)
    def test_default_schema_url_content(self):
        """Test the content of the remote URLs a bit. We only assert them not being None here just in case the server
        returns zero-length files."""

        init(update_only=False)

        # We only assert that we have retrieved some non-empty files in this test. This is because we want to decouple
        # the maintenance of schema files in production with the library development. These files likely would change
        # more regularly than the library. For this reason, we also verify the default schema URLs are also valid https
        # links in ``test_default_schema_url_https``.

        # This test is in `test_config.py` not in `test_schema_retrieval.py` because this test is more about the content
        # of the default schema URLs than the retrieving functionality.
        assert len(retrieve_schema_file(Config.DATASET_SCHEMA_URL)) > 0
        assert len(retrieve_schema_file(Config.FORMAT_SCHEMA_URL)) > 0
        assert len(retrieve_schema_file(Config.LICENSE_SCHEMA_URL)) > 0



# Copyright 2021 Edward Leardi. All Rights Reserved.
#
# Copyright 2020 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"Format to loader map."


from typing import Any, Dict, Mapping, Optional, Union

from .._schema import SchemaDict

from ._base import Loader
from .text import PlainTextLoader
from .table import CSVPandasLoader


class FormatLoaderMap:
    """Manage a map between formats and loaders. This is usually used to determine which loader should be used for a
    given format.

    :param m: A dict that maps formats to loaders.
    """

    def __init__(self, m: Optional[Mapping[str, Loader]] = None) -> None:
        """Constructor method.
        """
        self._map: Dict[str, Loader] = {}
        if m is not None:
            for fmt, loader in m.items():
                self.register_loader(fmt, loader)

    def register_loader(self, fmt: str, loader: Loader) -> None:
        """Register a loader. If the format exists in the table, update it.

        :param fmt: The format.
        :param loader: The corresponding loader.
        :raise TypeError: ``loader`` is not a :class:`Loader` object.
        """
        if not isinstance(loader, Loader):
            raise TypeError(f'loader "{loader}" must be a Loader instance.')
        # We may support an overriding check in the future
        self._map[fmt] = loader

    def __getitem__(self, fmt: str) -> Loader:
        """Get the loader of a given format.

        :param fmt: The format.
        """
        return self._map[fmt]

    def __contains__(self, fmt: str) -> bool:
        """Whether a format is covered by this format loader map.

        :param fmt: Name of the format.
        """
        return fmt in self._map


_default_format_loader_map: FormatLoaderMap = FormatLoaderMap({
    'txt': PlainTextLoader(),
    'csv': CSVPandasLoader()
})


def load_data_files(fmt: Union[str, SchemaDict], path: Union[str, Dict[str, str]], *,
                    format_loader_map: FormatLoaderMap = None) -> Any:
    """Load data files.

    :param fmt: The format.
    :param path: Path to the file(s).
    :param format_loader_map: The format loader map to use.
    :raises TypeError: ``fmt`` is neither a string nor a dict.
    :return: Loaded data file objects.
    """

    # We only support path as a plain path for now, but we will extend path to support regex and other types.

    if format_loader_map is None:
        format_loader_map = _default_format_loader_map

    if isinstance(fmt, str):
        fmt_id: str = fmt
        fmt_options: SchemaDict = {}
    elif isinstance(fmt, Dict):
        # In Python 3.8, this can be done with isinstance(fmt, typing.get_args(SchemaDict))
        fmt_id = fmt['id']
        fmt_options = fmt.get('options', {})
    else:
        raise TypeError(f'Parameter "fmt" must be a string or a dict, but it is of type "{type(fmt)}".')

    if fmt_id not in format_loader_map:
        raise RuntimeError(f'The format loader map does not specify a loader for format "{fmt_id}".')

    return format_loader_map[fmt_id].load(path, fmt_options)



# Copyright 2021 Edward Leardi. All Rights Reserved.
#
# Copyright 2020 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"Text file loaders."


import pathlib
from typing import cast, Dict, Union

from .. import typing as typing_
from ..schema import SchemaDict
from ._base import Loader


class PlainTextLoader(Loader):
    """Plain text to string loader.
    """

    def load(self, path: Union[typing_.PathLike, Dict[str, str]], options: SchemaDict) -> str:
        """The type hint says Dict, because this loader will be handling those situations in the future.

        :param path: The path to the plain text file.
        :param options:
               - ``encoding`` key specifies the encoding of the plain text.
        :raises TypeError: ``path`` is not a path object.
        :return: Data loaded into a ``str``.
        """

        super().load(path, options)

        encoding = options.get('encoding', 'utf-8')
        # We can remove usage of cast once Dict[str, str] handling is added
        path = cast(typing_.PathLike, path)
        return pathlib.Path(path).read_text(encoding=encoding)



# Copyright 2021 Edward Leardi. All Rights Reserved.
#
# Copyright 2020 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"Module for defining and modifying global configs."


import os
import pathlib
from typing import Union

from pydantic.dataclasses import dataclass


@dataclass(frozen=True)
class Config:
    """Global read-only configurations for Nourish.
    """

    # Default schema URLs
    # TODO: The types below should be typing_.PathLike. However, pydantic does not play well with os.PathLike for
    # validation. Will have to fix it in another occasion.
    DATASET_SCHEMA_URL: Union[str, pathlib.Path] = \
        'https://raw.githubusercontent.com/edwardleardi/nourish-schemata/main/datasets.yaml'
    FORMAT_SCHEMA_URL: Union[str, pathlib.Path] = \
        'https://raw.githubusercontent.com/edwardleardi/nourish-schemata/main/formats.yaml'
    LICENSE_SCHEMA_URL: Union[str, pathlib.Path] = \
        'https://raw.githubusercontent.com/edwardleardi/nourish-schemata/main/licenses.yaml'

    # DATADIR is the default dir where datasets files are downloaded/loaded to/from.
    DATADIR: pathlib.Path = pathlib.Path.home() / '.nourish' / 'data'

    def __post_init_post_parse__(self) -> None:
        "This is called by :meth:`.__init__()` after data type validation."
        # DATADIR should be absolute.
        # We use object.__setattr__ because we set frozen=True, same as what dataclasses does:
        # https://docs.python.org/3/library/dataclasses.html#frozen-instances
        object.__setattr__(self, 'DATADIR', pathlib.Path(os.path.abspath(self.DATADIR)))



# Copyright 2021 Edward Leardi. All Rights Reserved.
#
# Copyright 2020 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"Nourish package and high level functions."


from . import dataset, exceptions, loaders, schema
from ._high_level import (describe_dataset,
                          export_schemata,
                          get_config,
                          get_dataset_metadata,
                          init,
                          list_all_datasets,
                          load_dataset,
                          load_schemata)
from ._version import version as __version__

__all__ = (
           # modules & subpackages
           'dataset',
           'exceptions',
           'loaders',
           'schema',
           # high-level functions
           'describe_dataset',
           'export_schemata',
           'get_config',
           'get_dataset_metadata',
           'init',
           'list_all_datasets',
           'load_dataset',
           'load_schemata',
           # _version
           '__version__'
)



# Copyright 2021 Edward Leardi. All Rights Reserved.
#
# Copyright 2020 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"Retrieve remote schema file."


from pathlib import Path
import re
from typing import Union
from urllib.parse import urlparse
from urllib.request import urlopen

import requests
import requests.exceptions

from . import typing as typing_
from .exceptions import InsecureConnectionError


# Semantically, typing_.PathLike doesn't cover strings that represent URLs
def retrieve_schema_file(url_or_path: Union[typing_.PathLike, str], *,
                         encoding: str = 'utf-8',
                         tls_verification: Union[bool, typing_.PathLike] = True) -> str:
    """Retrieve a single schema file.

    :param url_or_path: URL or path to the schema file.
    :param encoding: The encoding of the text in ``url_or_path``.
    :param tls_verification: Same as ``tls_verification`` in :class:`nourish.Schema`.
    :raises ValueError: See :class:`nourish.Schema`.
    :raises InsecureConnectionError: See :class:`nourish.Schema`.
    :return: A string of the content.
    """

    url_or_path = str(url_or_path)

    # We don't detect fully whether the input is a URL or a file path because I couldn't find a reliable way. Almost any
    # string with no backslash can be a file name on Linux. URL detection often involves either giant dependencies such
    # as Django, or tediously long regular expression that we can't assure that it would work. Here, we detect the
    # beginning of the string. If it doesn't look like a URL, treat it as a file path.
    if re.match(r'[a-zA-Z0-9]+:\/\/', url_or_path):
        parse_result = urlparse(url_or_path)
        scheme = parse_result.scheme
        if scheme in ('http', 'https'):
            if scheme == 'http' and tls_verification:
                raise InsecureConnectionError((f'{url_or_path} is a http link and insecure. '
                                               'Set tls_verification=False to accept http links.'))
            try:
                content = requests.get(url_or_path, allow_redirects=True, verify=tls_verification).content
            except requests.exceptions.SSLError as e:
                raise InsecureConnectionError((f'Failed to securely connect to {url_or_path}. Caused by:\n{e}'))

            # We don't use requests.Response.encoding and requests.Response.text because it is always silent when
            # there's an encoding error
            return content.decode(encoding)
        elif scheme == 'file':
            with urlopen(url_or_path) as f:  # nosec: bandit will always complain but we know it points to a local file
                return f.read().decode(encoding)
        else:
            raise ValueError(f'Unknown scheme in "{url_or_path}": "{scheme}"')
    else:
        # Not a URL, treated as a local file path
        return Path(url_or_path).read_text(encoding)



#!/usr/bin/env python3

"""For doctests: We put a link in "tests" that points to "nourish". We make the symlink in "tests/doctests" instead
of directly under "tests" to avoid "tests/nourish" being imported during the tests and coverage won't be able to detect
coverage. This script is executed in tox.ini and should only be run from the project root."""

from pathlib import Path


path = Path("tests/doctests/nourish")

# We always try to clean up "tests/doctests/nourish" in case an outdated symlink is there (e.g., source directory being
# moved).
if path.exists():
    path.unlink()

path.parent.mkdir(exist_ok=True)
# We use a symlink to the absolute path to avoid potential system compatibility issues.
path.symlink_to(Path.cwd() / "nourish", target_is_directory=True)



# Copyright 2021 Edward Leardi. All Rights Reserved.
#
# Copyright 2020 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"Tabular data loaders."


from typing import Dict, Union

import pandas as pd  # type: ignore[import]

from .. import typing as typing_
from ..schema import SchemaDict
from ._base import Loader


class CSVPandasLoader(Loader):
    """CSV to Pandas dataframe loader.
    """

    def load(self, path: Union[typing_.PathLike, Dict[str, str]], options: SchemaDict) -> pd.DataFrame:
        """The type hint says Dict, because this loader will be handling those situations in the future, perhaps via a
        ``IteratingLoader`` class.

        :param path: The path to the CSV file.
        :param options:
               - ``columns`` key specifies the data type of each column. Each data type corresponds to a Pandas'
                 supported dtype. If unspecified, then it is default.
               - ``delimiter`` key specifies the delimiter of the input CSV file.
               - ``no_header`` key specifies if the first row of the CSV file contains the headers. Defaults to False.
                 If the value is set to anything "truthy" in Python, the first row of the CSV will be read as data.
               - ``encoding`` key specifies the encoding of the CSV file. Defaults to UTF-8.
        :raises TypeError: ``path`` is not a path object.
        :return: Data loaded into a :class:`pandas.DataFrame`.
        """

        super().load(path, options)

        parse_dates = []
        dtypes = {}
        for column, type_ in options.get('columns', {}).items():
            if type_ == 'datetime':
                # pandas has this unusual handling of date datatype. Instead of specifying as a data type of a column,
                # we have to pass in `parse_dates`.
                parse_dates.append(column)
            else:
                dtypes[column] = type_

        names = None
        header = None
        if options.get('no_header'):
            # If no header use the columns provided in schema
            names = [*options.get('columns', {})]
        else:
            header = 'infer'

        return pd.read_csv(path, dtype=dtypes,
                           # The following line after "if" is for circumventing
                           # https://github.com/pandas-dev/pandas/issues/38489
                           # TODO: Remove the if-else part of the following line once pandas has released version >1.1.5
                           # The bug above has been fixed.
                           parse_dates=parse_dates if len(parse_dates) > 0 else False,
                           header=header, names=names,
                           encoding=options.get('encoding', 'utf-8'),
                           delimiter=options.get('delimiter', ','))



# Copyright 2021 Edward Leardi. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"Dataset unarchiving and uncompressing functionality."


from abc import ABC, abstractmethod
import bz2
import gzip
import json
import lzma
import mimetypes
import pathlib
import shutil
import tarfile
from typing import Dict, Union
import zipfile


class Extractor(ABC):
    """Abstract class that provides functionality to extract dataset downloads.
    """

    @abstractmethod
    def extract(self, path: pathlib.Path, data_dir: pathlib.Path, file_list_file: pathlib.Path) -> None:
        """Extracts dataset download by unarchiving and/or uncompressing the files. This must be overridden when inherited.

        :param path: Path to the dataset to extract.
        :param data_dir: Path to the data dir to extract data files to.
        :file_list_file: Path to the file that stores the list of files in the downloaded dataset.
        """
        pass

    @abstractmethod
    def verify_extraction(self, data_dir: pathlib.Path, contents: dict) -> bool:
        """Verifies file extraction. This must be overridden when inherited.

        :param data_dir: Path to the data dir containing the extracted files.
        :contents: Dict obtained from ``file_list_file`` with metadata of files in the downloaded dataset.
        :return: ``True`` if the dataset was extracted properly and ``False`` otherwise.
        """
        pass


class _TarExtractor(Extractor):
    """Extractor that handles tarballs. Capable of handling files like: ``.tar``, ``.tar.gz``, ``.tar.bz2``,
    and ``.tar.xz``.
    """

    def extract(self, path: pathlib.Path, data_dir: pathlib.Path, file_list_file: pathlib.Path) -> None:
        """Attempt to extract the tar archive. Save metadata about the list of files in the downloaded dataset in
        ``file_list_file``.

        :param path: Path to the tar archive.
        :param data_dir: Path to the data dir to extract data files to.
        :file_list_file: Path to the file that stores the list of files in the downloaded dataset.
        :raises tarfile.ReadError: The tar archive was unable to be read.
        """
        try:
            mytar = tarfile.open(path)
        except tarfile.ReadError as e:
            raise tarfile.ReadError(f'Failed to unarchive tar file "{path}"\ncaused by:\n{e}')
        with mytar:
            FileListFileContents = Dict[str, Dict[str, int]]
            contents: FileListFileContents = {}
            metadata: Dict[str, Union[str, FileListFileContents]] = {}

            metadata['type'] = 'application/x-tar'
            for member in mytar.getmembers():
                contents[member.name] = {'type': int(member.type)}
                if member.isreg():  # For regular files, we also save its size
                    contents[member.name]['size'] = member.size
            metadata['contents'] = contents

            with open(file_list_file, mode='w') as f:
                # We do not specify 'utf-8' here to match the default encoding used by the OS, which also likely
                # uses this encoding for accessing the filesystem.
                json.dump(metadata, f, indent=2)
            mytar.extractall(path=data_dir)

    def verify_extraction(self, data_dir: pathlib.Path, contents: Dict[str, Dict[str, int]]) -> bool:
        """Verify the files were extracted properly using the metadata saved in ``file_list_file``.

        :param data_dir: Path to the data dir containing the extracted files.
        :contents: Dict obtained from ``file_list_file`` with metadata of files in the downloaded dataset.
        :return: ``True`` if the dataset was extracted properly and ``False`` otherwise.
        """
        for name, info in contents.items():
            path = data_dir / name
            if not path.exists():
                # At least one file in the file list is missing
                return False
            # We don't have pathlib type code that matches tarfile type code. We instead do an incomplete list of
            # type comparison. We don't do uncommon types such as FIFO, character device, etc. here.
            if info['type'] == int(tarfile.REGTYPE):  # Regular file
                if not path.is_file():
                    return False
                if path.stat().st_size != info['size']:
                    return False
            elif info['type'] == int(tarfile.DIRTYPE) and not path.is_dir():  # Directory type
                return False
            elif info['type'] == int(tarfile.SYMTYPE) and not path.is_symlink():  # Symbolic link type
                return False
            else:
                # We just let go any file types that we don't understand.
                pass
        return True


class _ZipExtractor(Extractor):
    """Extractor that handles zip archives. Capable of handling files like: ``.zip``.
    """

    def extract(self, path: pathlib.Path, data_dir: pathlib.Path, file_list_file: pathlib.Path) -> None:
        """Attempt to extract the zip archive. Save metadata about the list of files in the downloaded dataset in
        ``file_list_file``.

        :param path: Path to the zip archive.
        :param data_dir: Path to the data dir to extract data files to.
        :file_list_file: Path to the file that stores the list of files in the downloaded dataset.
        :raises zipfile.BadZipFile: The zip archive was unable to be read.
        """
        try:
            myzip = zipfile.ZipFile(path)
        except zipfile.BadZipFile as e:
            raise zipfile.BadZipFile(f'Failed to unarchive zip file "{path}"\ncaused by:\n{e}')
        with myzip:
            FileListFileContents = Dict[str, Dict[str, Union[bool, int]]]
            contents: FileListFileContents = {}
            metadata: Dict[str, Union[str, FileListFileContents]] = {}

            metadata['type'] = 'application/zip'
            for member in myzip.infolist():
                contents[member.filename] = {'isdir': member.is_dir()}
                if not member.is_dir():
                    contents[member.filename]['size'] = member.file_size
            metadata['contents'] = contents

            with open(file_list_file, mode='w') as f:
                json.dump(metadata, f, indent=2)
            myzip.extractall(path=data_dir)

    def verify_extraction(self, data_dir: pathlib.Path, contents: Dict[str, Dict[str, Union[bool, int]]]) -> bool:
        """Verify the files were extracted properly using the metadata saved in ``file_list_file``.

        :param data_dir: Path to the data dir containing the extracted files.
        :contents: Dict obtained from ``file_list_file`` with metadata of files in the downloaded dataset.
        :return: ``True`` if the dataset was extracted properly and ``False`` otherwise.
        """
        for name, info in contents.items():
            path = data_dir / name
            if not path.exists():
                # At least one file in the file list is missing
                return False
            # Ignore checking symlinks
            if not path.is_symlink():
                if path.is_dir() != info['isdir']:
                    return False
                if not info['isdir'] and (path.stat().st_size != info['size']):
                    return False
        return True


# gzip.BadGzipFile wasn't added until Python 3.8, so we use our own exception
class _BadGzipFile(Exception):
    "Not a gzip file."
    pass


class _GzipExtractor(Extractor):
    """Extractor that handles gzip compressed files. Capable of handling files like: ``.gz``, ``.txt.gz``, and
    ``.csv.gz``.
    """

    def extract(self, path: pathlib.Path, data_dir: pathlib.Path, file_list_file: pathlib.Path) -> None:
        """Attempt to extract the gzip compressed file. Save metadata about the file in the downloaded dataset in
        ``file_list_file``.

        :param path: Path to the gzip file.
        :param data_dir: Path to the data dir to extract the data file to.
        :file_list_file: Path to the file that stores metadata of the file in the downloaded dataset.
        :raises _BadGzipFile: The gzip file was unable to be read.
        :raises OSError: The extracted file is empty.
        """
        try:
            mygzip = gzip.open(path)
            extracted_file_name = path.stem
            with mygzip, open(data_dir / extracted_file_name, 'wb') as f_out:
                shutil.copyfileobj(mygzip, f_out)
        except OSError as e:
            raise _BadGzipFile(f'Failed to uncompress gzip file "{path}"\ncaused by:\n{e}')
        extracted_fp = data_dir / extracted_file_name

        FileListFileContents = Dict[str, Union[int, str]]
        contents: FileListFileContents = {}
        metadata: Dict[str, Union[str, FileListFileContents]] = {}

        metadata['type'] = 'gzip'
        contents['filename'] = extracted_file_name
        contents['size'] = extracted_fp.stat().st_size
        if contents['size'] == 0:
            raise OSError(f'The extracted file {extracted_file_name} is empty.')
        metadata['contents'] = contents

        with open(file_list_file, mode='w') as f:
            json.dump(metadata, f, indent=2)

    def verify_extraction(self, data_dir: pathlib.Path, contents: dict) -> bool:
        """Verify the file was extracted properly using the metadata saved in ``file_list_file``.

        :param data_dir: Path to the data dir containing the extracted file.
        :contents: Dict obtained from ``file_list_file`` with metadata of the file in the downloaded dataset.
        :return: ``True`` if the dataset was extracted properly and ``False`` otherwise.
        """
        path = data_dir / contents['filename']
        if not path.exists():
            # File is missing
            return False
        if path.stat().st_size != contents['size']:
            return False
        return True


class _BadBzip2File(Exception):
    "Not a bzip2 file."
    pass


class _Bzip2Extractor(Extractor):
    """Extractor that handles bzip2 compressed files. Capable of handling files like: ``.bz2``, ``.txt.bz2``, and
    ``.csv.bz2``.
    """

    def extract(self, path: pathlib.Path, data_dir: pathlib.Path, file_list_file: pathlib.Path) -> None:
        """Attempt to extract the bzip2 compressed file. Save metadata about the file in the downloaded dataset in
        ``file_list_file``.

        :param path: Path to the bzip2 file.
        :param data_dir: Path to the data dir to extract the data file to.
        :file_list_file: Path to the file that stores metadata of the file in the downloaded dataset.
        :raises _BadBzip2File: The bzip2 file was unable to be read.
        :raises OSError: The extracted file is empty.
        """
        try:
            mybzip2 = bz2.open(path)
            extracted_file_name = path.stem
            with mybzip2, open(data_dir / extracted_file_name, 'wb') as f_out:
                shutil.copyfileobj(mybzip2, f_out)
        except OSError as e:
            raise _BadBzip2File(f'Failed to uncompress bzip2 file "{path}"\ncaused by:\n{e}')
        extracted_fp = data_dir / extracted_file_name

        FileListFileContents = Dict[str, Union[int, str]]
        contents: FileListFileContents = {}
        metadata: Dict[str, Union[str, FileListFileContents]] = {}

        metadata['type'] = 'bzip2'
        contents['filename'] = extracted_file_name
        contents['size'] = extracted_fp.stat().st_size
        if contents['size'] == 0:
            raise OSError(f'The extracted file {extracted_file_name} is empty.')
        metadata['contents'] = contents

        with open(file_list_file, mode='w') as f:
            json.dump(metadata, f, indent=2)

    def verify_extraction(self, data_dir: pathlib.Path, contents: dict) -> bool:
        """Verify the file was extracted properly using the metadata saved in ``file_list_file``.

        :param data_dir: Path to the data dir containing the extracted file.
        :contents: Dict obtained from ``file_list_file`` with metadata of the file in the downloaded dataset.
        :return: ``True`` if the dataset was extracted properly and ``False`` otherwise.
        """
        path = data_dir / contents['filename']
        if not path.exists():
            # File is missing
            return False
        if path.stat().st_size != contents['size']:
            return False
        return True


class _LzmaExtractor(Extractor):
    """Extractor that handles lzma compressed files. Capable of handling files like: ``.xz``, ``.txt.xz``, and
    ``.csv.xz``.
    """

    def extract(self, path: pathlib.Path, data_dir: pathlib.Path, file_list_file: pathlib.Path) -> None:
        """Attempt to extract the lzma compressed file. Save metadata about the file in the downloaded dataset in
        ``file_list_file``.

        :param path: Path to the lzma file.
        :param data_dir: Path to the data dir to extract the data file to.
        :file_list_file: Path to the file that stores metadata of the file in the downloaded dataset.
        :raises lzma.LZMAError: The lzma file was unable to be read.
        :raises OSError: The extracted file is empty.
        """
        try:
            mylzma = lzma.open(path)
            extracted_file_name = path.stem
            with mylzma, open(data_dir / extracted_file_name, 'wb') as f_out:
                shutil.copyfileobj(mylzma, f_out)
        except lzma.LZMAError as e:
            raise lzma.LZMAError(f'Failed to uncompress lzma file "{path}"\ncaused by:\n{e}')
        extracted_fp = data_dir / extracted_file_name

        FileListFileContents = Dict[str, Union[int, str]]
        contents: FileListFileContents = {}
        metadata: Dict[str, Union[str, FileListFileContents]] = {}

        metadata['type'] = 'xz'
        contents['filename'] = extracted_file_name
        contents['size'] = extracted_fp.stat().st_size
        if contents['size'] == 0:
            raise OSError(f'The extracted file {extracted_file_name} is empty.')
        metadata['contents'] = contents

        with open(file_list_file, mode='w') as f:
            json.dump(metadata, f, indent=2)

    def verify_extraction(self, data_dir: pathlib.Path, contents: dict) -> bool:
        """Verify the file was extracted properly using the metadata saved in ``file_list_file``.

        :param data_dir: Path to the data dir containing the extracted file.
        :contents: Dict obtained from ``file_list_file`` with metadata of the file in the downloaded dataset.
        :return: ``True`` if the dataset was extracted properly and ``False`` otherwise.
        """
        path = data_dir / contents['filename']
        if not path.exists():
            # File is missing
            return False
        if path.stat().st_size != contents['size']:
            return False
        return True


# Keys based on ftype returned by mimetypes.guess_type
archives = {
    'application/x-tar': _TarExtractor(),
    'application/zip': _ZipExtractor()
}

# Keys based on fencoding returned by mimetypes.guess_type
compressions = {
    'gzip': _GzipExtractor(),
    'bzip2': _Bzip2Extractor(),
    'xz': _LzmaExtractor()
}

extractor_map = {**archives, **compressions}
read_errors = (tarfile.ReadError, zipfile.BadZipFile, _BadGzipFile, _BadBzip2File, lzma.LZMAError)


# We first run the extractor determined by guess_type
# If that returns an error or guess_type doesn’t know the type try all extractors
# If none of the extractors work raise an error that says file type unsupported
def extract_data_files(path: pathlib.Path, data_dir: pathlib.Path, file_list_file: pathlib.Path) -> None:
    """Choose extractor based on filetype and extract the data files.

    :param path: Path to the dataset to extract.
    :param data_dir: Path to the data dir to extract the data file to.
    :param file_list_file: Path to the file that stores the list of files in the downloaded dataset.
    :raises RuntimeError: Dataset filetype is not (yet) supported.
    """

    # mimetypes.guess_type doesn't accept path-like objects until python 3.8
    ftype, fencoding = mimetypes.guess_type(path.name)

    # 1. Check if mimetypes.guess_type guesses a working extractor
    if any(key in extractor_map.keys() for key in (ftype, fencoding)):
        # Check if file is an archive, otherwise file must be a compressed flat file
        if ftype in archives:
            extractor = archives[ftype]
        elif fencoding in compressions:
            extractor = compressions[fencoding]
        try:
            extractor.extract(path, data_dir, file_list_file)
        except read_errors:
            pass
        else:
            return

    # 2. Otherwise try all extractors and see if one works
    for extractor in extractor_map.values():
        try:
            extractor.extract(path, data_dir, file_list_file)
        except read_errors:
            pass
        else:
            return

    # 3. Otherwise assume unsupported flat file or compression/archive type
    raise RuntimeError('Filetype not (yet) supported')


def verify_data_files(data_dir: pathlib.Path, file_list_file: pathlib.Path) -> bool:
    """Verify the files were extracted properly using the metadata saved in ``file_list_file``.

    :param data_dir: Path to the data dir containing the extracted files.
    :param file_list_file: Path to the file containing the list of files in the downloaded dataset.
    :return: ``True`` if the dataset was extracted properly and ``False`` otherwise.
    """

    with open(file_list_file, mode='r') as f:
        metadata = json.load(f)
    extractor = extractor_map[metadata['type']]
    return extractor.verify_extraction(data_dir, metadata['contents'])



# Copyright 2021 Edward Leardi. All Rights Reserved.
#
# Copyright 2020 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"Script for sending a message to our gitter channel."

# To update token, go to https://developer.gitter.im/apps and copy token to the github repository secret GITTER_TOKEN
# To update room, find Gitter room id using Gitter API and copy id to the github repository secret GITTER_ROOM_ID

import os
import requests


GITTER_TOKEN = os.getenv('token')
GITTER_ROOM_ID = os.getenv('room-id')
GITTER_TEXT = os.getenv('text')
GIT_REF = os.getenv('GITHUB_REF')
GIT_SHA = os.getenv('GITHUB_SHA')
GITHUB_REPOSITORY = os.getenv('GITHUB_REPOSITORY')
GITHUB_SERVER_URL = os.getenv('GITHUB_SERVER_URL')
GITHUB_RUN_ID = os.getenv('GITHUB_RUN_ID')

headers = {
    'Content-Type': 'application/json',
    'Accept': 'application/json',
    'Authorization': f"Bearer {GITTER_TOKEN}",
}

data = (
    f'{{"text":"**{GITHUB_REPOSITORY}**\\n\\n'
    f'{GITTER_TEXT}\\n'
    f'{GIT_REF}:{GIT_SHA}\\n'
    f'{GITHUB_SERVER_URL}/{GITHUB_REPOSITORY}/actions/runs/{GITHUB_RUN_ID}"}}'
)

response = requests.post(f'https://api.gitter.im/v1/rooms/{GITTER_ROOM_ID}/chatMessages', headers=headers, data=data)



# Copyright 2021 Edward Leardi. All Rights Reserved.
#
# Copyright 2020 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This file has been modified by Edward Leardi to support downloading & extracting new dataset formats
#

"Dataset downloading and loading functionality."


from enum import IntFlag
import hashlib
import os
import pathlib
import shutil
from typing import Any, Callable, Dict, Iterable, Optional

import requests

from . import typing as typing_
from .loaders import FormatLoaderMap
from .loaders._format_loader_map import load_data_files
from .schema import SchemaDict
from ._extractors import extract_data_files, verify_data_files
from ._lock import DirectoryLock


class Dataset:
    """Models a particular dataset version along with download & load functionality.

    :param schema: Schema dict of a particular dataset version.
    :param data_dir: Directory to/from which the dataset should be downloaded/loaded from. The path can be either
        absolute or relative to the current working directory, but will be converted to the absolute path immediately
        upon initialization.
    :param mode: Mode with which to treat a dataset. Available options are:
        :attr:`Dataset.InitializationMode.LAZY`, :attr:`Dataset.InitializationMode.DOWNLOAD_ONLY`,
        :attr:`Dataset.InitializationMode.LOAD_ONLY`, and :attr:`Dataset.InitializationMode.DOWNLOAD_AND_LOAD`.
    :raises ValueError: An invalid ``mode`` was specified for handling the dataset.

    Example:

    >>> from tempfile import TemporaryDirectory
    >>> import nourish
    >>> from nourish import schema
    >>> dataset_schema = schema.DatasetSchema('./tests/schemata/datasets.yaml')
    >>> jfk_schema_dict = dataset_schema.export_schema('datasets', 'noaa_jfk', '1.1.4')
    >>> jfk_data_dir = TemporaryDirectory()
    >>> jfk_dataset = Dataset(schema=jfk_schema_dict, data_dir=jfk_data_dir.name)
    >>> jfk_dataset.download()
    >>> data = jfk_dataset.load()
    >>> data['jfk_weather_cleaned'].shape
    (75119, 16)
    >>> jfk_dataset.delete()  # The directory jfk_data_dir is deleted here
    >>> jfk_dataset.is_downloaded()
    False
    """

    # Note: we use a convention for attribute pairs that return the same pathlib.Path where a trailing "_" indicates
    # the path should be used for reading data (e.g. checking the path exists) while no trailing "_" indicates the path
    # is for writing data (e.g. creating the file)

    class InitializationMode(IntFlag):
        """Enum class that acts as `mode` for :class:`Dataset`.
        """
        LAZY = 0
        DOWNLOAD_ONLY = 1
        LOAD_ONLY = 2
        DOWNLOAD_AND_LOAD = 3

    def __init__(self, schema: SchemaDict,
                 data_dir: typing_.PathLike,  *,
                 mode: InitializationMode = InitializationMode.LAZY) -> None:
        """Constructor method.
        """

        self._schema: SchemaDict = schema
        self._data_dir_: pathlib.Path = pathlib.Path(os.path.abspath(data_dir))
        self._data: Optional[Dict[str, Any]] = None
        # Put directory lock under self._nourish_dir. We use self._nourish_dir_ instead of self._nourish_dir because we
        # don't want to have the directory created in lazy mode upon construction of a Dataset object.
        self._lock: DirectoryLock = DirectoryLock(self._nourish_dir_)

        if not isinstance(mode, Dataset.InitializationMode):
            raise ValueError(f'{mode} not a valid mode')

        if mode & Dataset.InitializationMode.DOWNLOAD_ONLY:
            self.download()
        if mode & Dataset.InitializationMode.LOAD_ONLY:
            self.load()

    @property
    def _data_dir(self) -> pathlib.Path:
        "Same as :attr:`_data_dir_`, but create it if it does not exist."
        if not self._data_dir_.exists():
            self._data_dir_.mkdir(parents=True)
        elif not self._data_dir_.is_dir():  # self._data_dir_ exists and is not a directory
            raise NotADirectoryError(f'"{self._data_dir_}" exists and is not a directory.')
        return self._data_dir_

    @property
    def _nourish_dir_(self) -> pathlib.Path:
        "Cache, metainfo, etc. directory used by this class."
        return self._data_dir_ / '.nourish.dataset'

    @property
    def _nourish_dir(self) -> pathlib.Path:
        "Same as :attr:`_nourish_dir_`, but create it if it does not exist."
        if not self._nourish_dir_.exists():
            self._nourish_dir_.mkdir(parents=True)
        elif not self._nourish_dir_.is_dir():  # nourish_dir exists and is not a directory
            raise NotADirectoryError(f'"{self._nourish_dir_}" exists and is not a directory.')
        return self._nourish_dir_

    @property
    def _file_list_file_(self) -> pathlib.Path:
        "Path to the file that stores the list of files in the downloaded dataset."
        return self._nourish_dir_ / 'files.list'

    @property
    def _file_list_file(self) -> pathlib.Path:
        "Same as :attr:`_file_list_file_`, but create the parent directory if it does not exist."
        return self._nourish_dir / 'files.list'

    def download(self,
                 check: bool = True) -> None:
        """Downloads, extracts, and removes dataset archive. It adds a directory write lock during execution.

        :param check: Check to make sure the data files are not already present in :attr:`._data_dir` (passed in via
            ``data_dir`` in the constructor :class:`Dataset`) by running :meth:`.is_downloaded`. If set to ``True``,
            raise an error if they are present and prevent a subsequent download. Set to ``False`` to remove this
            safeguard, and subsequent calls to :meth:`.download` will then overwrite data files if they were previously
            downloaded to :attr:`._data_dir`.
        :raises RuntimeError: The dataset was previously downloaded as indicated by :meth:`.is_downloaded`
            returning ``True``.
        :raises NotADirectoryError: :attr:`Dataset._data_dir` (passed in via ``data_dir`` in the constructor
            :class:`Dataset`) points to an existing file that is not a directory.
        :raises OSError: The SHA512 checksum of a downloaded dataset doesn't match the expected checksum.
        :raises exceptions.DirectoryLockAcquisitionError: Failed to acquire the directory lock.
        """

        if check and self.is_downloaded():
            raise RuntimeError(f'{self.__class__.__name__}.download() was previously called. To overwrite existing '
                               f'data files, rerun {self.__class__.__name__}.download() with ``check`` set to '
                               f'``False``.')

        download_url = self._schema['download_url']
        download_file_name = pathlib.Path(os.path.basename(download_url))

        with self._lock.locking_with_exception(write=True):
            archive_fp = self._nourish_dir / download_file_name
            response = requests.get(download_url, stream=True)
            archive_fp.write_bytes(response.content)

            computed_hash = hashlib.sha512(archive_fp.read_bytes()).hexdigest()
            actual_hash = self._schema['sha512sum']
            if not actual_hash == computed_hash:
                raise OSError(f'{archive_fp} has a SHA512 checksum of: ({computed_hash}) '
                              f'which is different from the expected SHA512 checksum of: ({actual_hash}) '
                              f'the file may by corrupted.')

            extract_data_files(path=archive_fp, data_dir=self._data_dir, file_list_file=self._file_list_file)
            os.remove(archive_fp)

    def load(self,
             subdatasets: Optional[Iterable[str]] = None,
             format_loader_map: Optional[FormatLoaderMap] = None,
             check: bool = True) -> Dict[str, Any]:
        """Load data files to RAM. It adds a directory read lock during execution.

        :param subdatasets: The subdatasets to load. ``None`` means all subdatasets.
        :param format_loader_map: The :class:`.loaders.FormatLoaderMap` object that determines which loader to use.
        :param check: Check to make sure the data files are currently present in :attr:`._data_dir` (passed in via
            ``data_dir`` in the constructor :class:`Dataset`) before loading them by running :meth:`.is_downloaded`.
            If set to ``True``, raise an error if they are missing and prevent attempting to load them. Set to ``False``
            to remove this safeguard.
        :raises RuntimeError: The dataset was not previously downloaded as indicated by :meth:`.is_downloaded`
            returning ``False``.
        :raises FileNotFoundError: The dataset files for a particular subdataset are not found on the disk. Either this
            is because :func:`~Dataset.download` was never called, or the dataset was only partially downloaded.
        :raises exceptions.DirectoryLockAcquisitionError: Failed to acquire the directory lock.
        :return: Loaded data objects. Same as :attr:`.data`.
        """

        if subdatasets is None:
            subdatasets = self._schema['subdatasets'].keys()

        if check and not self.is_downloaded():
            raise RuntimeError(f'Downloaded data files are not present in {self._data_dir_} or are corrupted.')

        with self._lock.locking_with_exception(write=False):
            self._data = {}
            for subdataset in subdatasets:
                subdataset_schema = self._schema['subdatasets'][subdataset]
                try:
                    self._data[subdataset] = load_data_files(fmt=subdataset_schema['format'],
                                                             path=self._data_dir / subdataset_schema['path'],
                                                             format_loader_map=format_loader_map)
                except FileNotFoundError as e:
                    self._data = None
                    raise FileNotFoundError(
                        f'Failed to load subdataset "{subdataset}" because some files are not found. '
                        f'Did you forget to call {self.__class__.__name__}.download()?\nCaused by:\n{e}')

        return self.data

    def delete(self, *, force: bool = False) -> None:
        """Clear the data directory. It adds a directory write lock before deletion during execution if the data
        directory exists.

        :param force: If ``True``, delete the directory even if directory locks are present.
        :raises exceptions.DirectoryLockAcquisitionError: Failed to acquire the directory lock.
        """

        if self._data_dir_.exists():
            if force:
                lock_func: Callable = self._lock.locking
            else:
                lock_func = self._lock.locking_with_exception
            with lock_func(write=True):
                shutil.rmtree(self._data_dir_)

    @property
    def data(self) -> Dict[str, Any]:
        """Access loaded data objects."""
        if self._data is None:
            raise RuntimeError(f'Data has not been downloaded and/or loaded yet. Call '
                               f'{self.__class__.__name__}.download() to download data, call '
                               f'{self.__class__.__name__}.load() to load data.')
        # We don't copy here because it is too expensive and users may actually want to update the datasets and it
        # doesn't cause security issues as in the Schema class
        return self._data

    def is_downloaded(self) -> bool:
        """Check to see if the dataset was downloaded. We determine this by comparing the extracted file tree with the file
        list :meth:`._file_list_file` (their existence, types, and sizes). In this way, if the extraction of the archive
        failed, this should return ``False`` and the user would not be misled. For performance reasons, we do not
        examine the content of the extracted files.

        :return: ``True`` if the dataset has been downloaded and ``False`` otherwise.

        .. warning::

            :meth:`.is_downloaded` will search for the dataset files in :attr:`Dataset._data_dir` (passed in via
            ``data_dir`` in the constructor :class:`Dataset`). If after downloading, you manipulate the data files
            outside the control of this library, this method may produce unexpected behavior.
        """

        # The method to detect whether the dataset has been downloaded can certainly be improved by balancing how much
        # to examine and how much time it takes to do so, considering some heuristics such as the dataset size, the
        # number of files, etc. The method used here should be able to strike a good balance for most cases and should
        # be good enough for the first release.

        if not self._file_list_file_.exists():
            # File not found, may not have finished downloading at all and we treat it as so. We can't control users'
            # own tweaking with the directory.
            return False
        return verify_data_files(data_dir=self._data_dir, file_list_file=self._file_list_file_)



# Copyright 2021 Edward Leardi. All Rights Reserved.
#
# Copyright 2020 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"Loader base class."


from abc import ABC, abstractmethod
import os
from typing import Any, Dict, Union

from .. import typing as typing_
from .._schema import SchemaDict


class Loader(ABC):
    """Base class of all loaders.
    """

    @abstractmethod
    def load(self, path: Union[typing_.PathLike, Dict[str, str]], options: SchemaDict) -> Any:
        """Loads from a given path or a dict of path configurations. This must be overridden when inherited.

        :param path: The path or path configurations of the files to be loaded.
        :param options: Options passed to the loader.
        :return: The object representing the loaded file.
        """
        self.check_path(path)

    def check_path(self, path: Union[typing_.PathLike, Dict[str, str]]) -> None:
        """Check if the given path is a valid path to the file to be loaded. Raise an error if it is not.

        :param path: The path of the file to be loaded.
        :raises TypeError: ``path`` is not a path object.
        :return: No return unless the ``path`` is invalid, in which case see :class:`TypeError`.
        """
        if not isinstance(path, (str, os.PathLike)):
            # In Python 3.8, this can be done with isinstance(path, typing.get_args(typing_.PathLike))
            raise TypeError(f'Unsupported path type "{type(path)}".')



# -*- coding: utf-8 -*-
#
# Copyright 2021 Edward Leardi. All Rights Reserved.
#
# Copyright 2020 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import sphinx_rtd_theme  # noqa: F401

from nourish import __version__ as nourish_version

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#

# sys.path.insert(0, os.path.abspath('../..'))


# -- Project information -----------------------------------------------------

project = 'Nourish'
copyright = '2020–2021, Edward Leardi'
version = nourish_version
release = version


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.intersphinx',
    'sphinx_rtd_theme'
    ]

autodoc_inherit_docstrings = True  # If no class summary, inherit base class summary
autodoc_typehints = 'description'  # Put parameter types to the description section to unclog prototypes
autosummary_generate = True  # Automate sphinx.ext.autosummary to run when html is generated
autosummary_imported_members = True  # Imported members should be documented
intersphinx_mapping = {
    'pandas': ('https://pandas.pydata.org/pandas-docs/stable/', None),
    'python': ('https://docs.python.org/3', None)
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
html_theme_options = {
    # TODO: add analytics_id to track traffic in Google Analytics
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# CSS files
html_css_files = ['css/custom.css']



# Copyright 2021 Edward Leardi. All Rights Reserved.
#
# Copyright 2020 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from collections import namedtuple
import re

import pytest
import pandas as pd
from pandas.api.types import is_datetime64_any_dtype, is_float_dtype, is_integer_dtype, is_string_dtype

from nourish.dataset import Dataset
from nourish.loaders import Loader
from nourish.loaders import FormatLoaderMap
from nourish.loaders._format_loader_map import load_data_files
from nourish.loaders.text import PlainTextLoader
from nourish.loaders.table import CSVPandasLoader


class TestBaseLoader:
    "Test loaders._base.*"

    def test_abstract(self):
        "Loader is an abstract class."

        with pytest.raises(TypeError) as e:
            Loader()
        assert 'abstract class' in str(e.value)

    def test_load(self, tmp_path):
        "Loader.load() must be overridden upon Loader being inherited."

        class MyLoader(Loader):
            pass

        # Error out when instantiating MyLoader because load method is not overridden
        with pytest.raises(TypeError) as e:
            MyLoader()
        assert "Can't instantiate abstract class MyLoader with abstract method" in str(e.value)

        class MyLoader(Loader):
            def load(self, path, options):
                # Calling the parent's load() method shouldn't lead to error
                super().load(path, options)
        # This line shouldn't error out even though it calls an abstract method in its parent
        MyLoader().load(tmp_path, None)

    def test_check_path(self):
        "Test Loader.check_path method."

        class MyLoader(Loader):
            def load(self):
                pass

        loader = MyLoader()
        integer = 1
        with pytest.raises(TypeError) as e:
            loader.check_path(integer)
        assert str(e.value) == f'Unsupported path type "{type(integer)}".'


class TestFormatLoaderMap:
    "Test loaders._format_loader.*"

    def test_register_non_loader(self):
        "Test when it registers a non-Loader instance."

        flm = FormatLoaderMap()

        with pytest.raises(TypeError) as e:
            flm.register_loader('some-format', 'some-string')
        assert str(e.value) == 'loader "some-string" must be a Loader instance.'

    def test_load_non_existing_format(self, tmp_path):
        "Test loading a non-existing format."

        with pytest.raises(RuntimeError) as e:
            load_data_files('nonsense', tmp_path)

        assert str(e.value) == 'The format loader map does not specify a loader for format "nonsense".'

    def test_load_wrong_format_type(self, tmp_path):
        "Test loading a non-existing format."

        with pytest.raises(TypeError) as e:
            load_data_files(0x348f, tmp_path)

        assert str(e.value) == 'Parameter "fmt" must be a string or a dict, but it is of type "<class \'int\'>".'


class TestTextLoaders:

    def test_plain_text_loader_no_path(self):
        "Test PlainTextLoader when fed in with non-path."

        integer = 1
        with pytest.raises(TypeError) as e:
            PlainTextLoader().load(integer, {})

        assert str(e.value) == f'Unsupported path type "{type(integer)}".'

    def test_plain_text_loader_bad_encoding(self, tmp_path):
        "Test PlainTextLoader when the encoding is nonsense."

        text_file = tmp_path / 'some-text.txt'
        text_file.write_text("I'm a text file :)", encoding='utf-8')
        with pytest.raises(LookupError) as e:
            PlainTextLoader().load(text_file, {'encoding': "non-encoding"})

        assert str(e.value) == 'unknown encoding: non-encoding'

    def test_plain_text_loader_incorrect_encoding(self, tmp_path):
        "Test PlainTextLoader when the encoding does not match."

        text_file = tmp_path / 'some-text.txt'
        text_file.write_text("I'm a text file :)", encoding='utf-8')
        with pytest.raises(UnicodeError) as e:
            PlainTextLoader().load(text_file, {'encoding': "utf-16"})

        assert str(e.value) == 'UTF-16 stream does not start with BOM'


class TestTableLoaders:

    def test_csv_pandas_loader(self, tmp_path, noaa_jfk_schema):
        "Test the basic functioning of CSVPandasLoader."

        dataset = Dataset(noaa_jfk_schema, tmp_path, mode=Dataset.InitializationMode.DOWNLOAD_AND_LOAD)
        data = dataset.data['jfk_weather_cleaned']
        assert isinstance(data, pd.DataFrame)
        assert data.shape == (75119, 16)
        dataset.delete()

    Column = namedtuple('Column', ('name', 'dtype', 'check'))

    @pytest.mark.parametrize('columns',  # a list of Column (column name, specified data type, check function)
                             [
                                 # Only one column specified
                                 [Column('DATE', 'datetime', is_datetime64_any_dtype)],
                                 [Column('DATE', 'str', is_string_dtype)],
                                 [Column('DATE', 'string', is_string_dtype)],
                                 [Column('HOURLYPressureTendencyCons', 'float', is_float_dtype)],
                                 # Two columns specified
                                 [Column('DATE', 'datetime', is_datetime64_any_dtype),
                                  Column('HOURLYPressureTendencyCons', 'float', is_float_dtype)],
                                 # No column specified (let Pandas autodetect dtype)
                                 [Column('DATE', None, is_string_dtype),
                                  Column('HOURLYPressureTendencyCons', None, is_integer_dtype),
                                  Column('HOURLYVISIBILITY', None, is_float_dtype)],
                             ])
    def test_csv_pandas_column_data_types(self, tmp_path, noaa_jfk_schema, columns):
        "Test the column data types."

        assert len(columns) > 0  # Sanity check, make sure columns are there

        # Clear columns
        column_dict = noaa_jfk_schema['subdatasets']['jfk_weather_cleaned']['format']['options']['columns'] = {}

        # Update column dictionary as specified
        for col in columns:
            if col.dtype is not None:
                column_dict[col.name] = col.dtype

        dataset = Dataset(noaa_jfk_schema, tmp_path, mode=Dataset.InitializationMode.DOWNLOAD_AND_LOAD)
        data = dataset.data['jfk_weather_cleaned']
        for col in columns:
            assert col.check(data.dtypes[col.name])

    @pytest.mark.parametrize(('err_column',  # (column name, specified data type, default dtype checked for conversion)
                              'other_columns'),  # (column name, specified data type, None)
                             [
                                 # Only one unsupported specified
                                 (Column('DATE', 'float', 'str'), []),
                                 (Column('HOURLYVISIBILITY', 'int', 'float'), []),
                                 # Some supported specified
                                 (Column('DATE', 'float', 'str'), [Column('HOURLYPressureTendencyCons', 'int', None)]),
                                 (Column('HOURLYVISIBILITY', 'int', 'float'), [Column('DATE', 'datetime', None)]),
                                 # More than one unsupported specified. The error that raises the exception should be
                                 # put as err_column.
                                 (Column('DATE', 'float', 'str'), [Column('HOURLYVISIBILITY', 'int', 'float')]),
                             ])
    def test_csv_pandas_column_unsupported_data_types(self, tmp_path, noaa_jfk_schema,
                                                      err_column, other_columns):
        "Test column data types when they are unsupported."

        # Clear columns
        column_dict = noaa_jfk_schema['subdatasets']['jfk_weather_cleaned']['format']['options']['columns'] = {}

        # Update column dictionary as specified
        for col in other_columns:
            if col.dtype is not None:
                column_dict[col.name] = col.dtype
        column_dict[err_column.name] = err_column.dtype

        with pytest.raises(ValueError) as e:
            Dataset(noaa_jfk_schema, tmp_path, mode=Dataset.InitializationMode.DOWNLOAD_AND_LOAD)
        # Pandas is a 3rd-party library. We don't check for the exact wording but only some keywords
        # Examples:
        #   ValueError: cannot safely convert passed user dtype of int64 for float64 dtyped data in column 1
        #   ValueError: could not convert string to float: '2010-01-01 01:00:00'
        assert 'convert' in str(e.value)
        for t in (err_column.dtype, err_column.check):
            assert re.search(rf"{t}(\d*|ing)\b", str(e.value))  # "ing" is for "str'ing'"

    def test_csv_pandas_no_delimiter(self, tmp_path, noaa_jfk_schema):
        "Test when no delimiter is given."
        # Remove the delimiter option
        del noaa_jfk_schema['subdatasets']['jfk_weather_cleaned']['format']['options']['delimiter']
        data = Dataset(noaa_jfk_schema, tmp_path,
                       mode=Dataset.InitializationMode.DOWNLOAD_AND_LOAD).data['jfk_weather_cleaned']
        assert len(data.columns) == 16  # number of columns remain the same

    @pytest.mark.parametrize('delimiter', ('\t', ' ', '|', ';'))
    def test_csv_pandas_delimiter(self, tmp_path, noaa_jfk_schema, delimiter):
        "Test common delimiter settings. Note that the case of comma has been tested in ``test_csv_pandas_loader``."

        del noaa_jfk_schema['subdatasets']['jfk_weather_cleaned']['format']['options']['columns']
        # Change delimiter to tab, |, ;, space
        noaa_jfk_schema['subdatasets']['jfk_weather_cleaned']['format']['options']['delimiter'] = delimiter
        data = Dataset(noaa_jfk_schema, tmp_path,
                       mode=Dataset.InitializationMode.DOWNLOAD_AND_LOAD).data['jfk_weather_cleaned']
        # None of these delimiters exist in the file, number of columns should be 1.
        assert len(data.columns) == 1

    def test_csv_pandas_loader_no_path(self):
        "Test CSVPandasLoader when fed in with non-path."

        integer = 1
        with pytest.raises(TypeError) as e:
            CSVPandasLoader().load(integer, {})

        assert str(e.value) == f'Unsupported path type "{type(integer)}".'

    def test_csv_pandas_loader_non_option(self, tmp_path, noaa_jfk_schema):
        "Test CSVPandasLoader when None option is passed."

        del noaa_jfk_schema['subdatasets']['jfk_weather_cleaned']['format']['options']
        dataset = Dataset(noaa_jfk_schema, tmp_path, mode=Dataset.InitializationMode.DOWNLOAD_AND_LOAD)
        data = dataset.data['jfk_weather_cleaned']
        assert isinstance(data, pd.DataFrame)
        assert len(data) == 75119

    def test_csv_pandas_loader_no_encoding(self, tmp_path, noaa_jfk_schema):
        "Test CSVPandasLoader when no encoding is specified."

        del noaa_jfk_schema['subdatasets']['jfk_weather_cleaned']['format']['options']['encoding']
        self.test_csv_pandas_loader(tmp_path, noaa_jfk_schema)

    def test_csv_pandas_header(self, tmp_path, noaa_jfk_schema):
        "Test CSVPandasLoader header options"

        noaa_jfk_schema['subdatasets']['jfk_weather_cleaned']['format']['options']['no_header'] = True
        noaa_dataset = Dataset(noaa_jfk_schema, tmp_path, mode=Dataset.InitializationMode.DOWNLOAD_ONLY)
        with pytest.raises(ValueError) as exinfo:  # Pandas should error from trying to read string as another dtype
            noaa_dataset.load()
        assert('could not convert string to float' in str(exinfo.value))
        noaa_dataset.delete()

        false_test_cases = [False, '', None]  # These should all be treated as False
        for case in false_test_cases:
            noaa_jfk_schema['subdatasets']['jfk_weather_cleaned']['format']['options']['no_header'] = case
            self.test_csv_pandas_loader(tmp_path, noaa_jfk_schema)

        del noaa_jfk_schema['subdatasets']['jfk_weather_cleaned']['format']['options']['no_header']
        self.test_csv_pandas_loader(tmp_path, noaa_jfk_schema)



# Copyright 2021 Edward Leardi. All Rights Reserved.
#
# Copyright 2020 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"Schema parsing and loading functionality."


from abc import ABC
from copy import deepcopy
from typing import Any, Dict, Union

import yaml

from . import typing as typing_
from ._schema_retrieval import retrieve_schema_file


SchemaDict = Dict[str, Any]


class Schema(ABC):
    """Abstract class that provides functionality to load and export schemata.

    :param url_or_path: URL or path to a schema file.
    :param tls_verification: When set to ``True``, verify the remote link is https and whether the TLS certificate is
        valid. When set to a path to a file, use this file as a CA bundle file. When set to ``False``, allow http links
        and do not verify any TLS certificates. Ignored if ``url_or_path`` is a local path.
    :raises ValueError: An error occurred when parsing ``url_or_path`` as either a URL or path.
    :raises InsecureConnectionError: The connection is insecure. See ``tls_verification`` for more details.
    """

    def __init__(self, url_or_path: Union[typing_.PathLike, str], *,
                 tls_verification: Union[bool, typing_.PathLike] = True) -> None:
        """Constructor method.
        """
        self._schema: SchemaDict = self._load_retrieved_schema(retrieve_schema_file(url_or_path,
                                                                                    tls_verification=tls_verification))

        # The URL or path from which the schema was retrieved
        self._retrieved_url_or_path: Union[typing_.PathLike, str] = url_or_path

    def _load_retrieved_schema(self, schema: str) -> SchemaDict:
        """Safely loads retrieved schema file.

        :param schema: Retrieved schema object.
        :return: Nested dictionary representation of a schema.
        """
        return yaml.safe_load(schema)

    def export_schema(self, *keys: str) -> SchemaDict:
        """Returns a copy of a loaded schema. Should be used for debug purposes only.

        :param keys: The sequence of keys that leads to the portion of the schema to be exported.
        :return: Copy of the schema dictionary.

        Example:

        >>> schema = DatasetSchema('./tests/schemata/datasets.yaml')
        >>> schema.export_schema('datasets', 'noaa_jfk', '1.1.4')
        {'name': 'NOAA Weather Data – JFK Airport'...}
        """
        schema: SchemaDict = self._schema
        for k in keys:
            schema = schema[k]
        return deepcopy(schema)

    @property
    def retrieved_url_or_path(self) -> Union[typing_.PathLike, str]:
        """The URL or path from which the schema was retrieved.

        Example:

        >>> schema = DatasetSchema('./tests/schemata/datasets.yaml')
        >>> schema.retrieved_url_or_path
        './tests/schemata/datasets.yaml'
        """
        return self._retrieved_url_or_path


class DatasetSchema(Schema):
    """Dataset schema class that inherits functionality from :class:`Schema`.
    """

    # We have this class here because we reserve the potential to put specific dataset schema code here
    pass


class FormatSchema(Schema):
    """Format schema class that inherits functionality from :class:`Schema`.
    """

    # We have this class here because we reserve the potential to put specific format schema code here
    pass


class LicenseSchema(Schema):
    """License schema class that inherits functionality from :class:`Schema`.
    """

    # We have this class here because we reserve the potential to put specific license schema code here
    pass


class SchemaManager():
    """Stores all loaded schemata in :attr:`schemata`.

    :param kwargs: Schema name and schema instance key-value pairs.

    Example:

    >>> dataset_schema = DatasetSchema('./tests/schemata/datasets.yaml')
    >>> schema_manager = SchemaManager(datasets=dataset_schema)
    >>> licenses_schema = LicenseSchema('./tests/schemata/licenses.yaml')
    >>> schema_manager.add_schema('licenses', licenses_schema)
    >>> schema_manager.schemata
    {'datasets':..., 'licenses':...}
    """

    def __init__(self, **kwargs: Schema) -> None:
        """Constructor method
        """
        self.schemata: Dict[str, Schema] = {}
        for name, val in kwargs.items():
            self.add_schema(name, val)

    def add_schema(self, name: str, val: Schema) -> None:
        """Store schema instance in a dictionary. If a schema with the same name as ``name`` is already stored, it is
        overridden.

        :param name: Schema name.
        :param val: Schema instance.
        """
        if not isinstance(val, Schema):
            raise TypeError('val must be a Schema instance.')
        self.schemata[name] = val



# Copyright 2021 Edward Leardi. All Rights Reserved.
#
# Copyright 2020 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os

import pytest

from nourish._lock import DirectoryLock
from nourish.exceptions import DirectoryLockAcquisitionError


class TestDirectoryLock:
    "Test :class:`DirectoryLock`."

    @staticmethod
    def _ensure_lock_unlock_succeeds(directory: os.PathLike, write: bool):
        "A utility function that ensures some locks and unlocks should succeed."
        lock = DirectoryLock(directory)
        prefix = 'write' if write else 'read'
        assert lock.lock(write=write) is True
        assert (directory / f'{prefix}{lock._lock_file_suffix}').exists()
        assert lock.unlock() is True
        assert not (directory / f'{prefix}{lock._lock_file_suffix}').exists()

        with lock.locking(write=write) as succeed:
            assert (directory / f'{prefix}{lock._lock_file_suffix}').exists()
            assert succeed is True
        assert not (directory / f'{prefix}{lock._lock_file_suffix}').exists()

        with lock.locking_with_exception(write=write):
            assert (directory / f'{prefix}{lock._lock_file_suffix}').exists()
        assert not (directory / f'{prefix}{lock._lock_file_suffix}').exists()

    @staticmethod
    def _ensure_lock_fails(directory: os.PathLike, write: bool):
        "A utility function that ensures some locks should fail."
        lock = DirectoryLock(directory)
        prefix = 'write' if write else 'read'
        assert lock.lock(write=write) is False
        assert not (directory / f'{prefix}{lock._lock_file_suffix}').exists()

        with lock.locking(write=write) as succeed:
            assert succeed is False
            assert not (directory / f'{prefix}{lock._lock_file_suffix}').exists()
        assert not (directory / f'{prefix}{lock._lock_file_suffix}').exists()

        with pytest.raises(DirectoryLockAcquisitionError) as e:
            with lock.locking_with_exception(write=write):
                pass
        assert str(e.value) == f'Failed to acquire directory {"write" if write else "read"} lock for "{directory}"'
        assert not (directory / f'{prefix}{lock._lock_file_suffix}').exists()

    @staticmethod
    def _ensure_unlock_fails(directory: os.PathLike):
        "A utility function that ensures some unlocks should fail."
        lock = DirectoryLock(directory)
        assert lock.unlock() is False

    def test_lock_empty(self, tmp_path):
        "Test lock in an empty directory. Unlocking without locking should also fail."
        for write in (True, False):
            self._ensure_lock_unlock_succeeds(tmp_path, write)

        self._ensure_unlock_fails(tmp_path)

    @pytest.mark.parametrize('lock_file_list',
                             [
                                 # One write lock
                                 ('write.ding.dong.lock',),
                                 # Two write locks
                                 ('write.dong.ding.lock', 'write.777.lock'),
                                 # One write lock and one read lock
                                 ('write.ding.dong.lock', 'read.bing-bong.lock'),
                             ])
    def test_failing_write_lock(self, tmp_path, lock_file_list):
        "Test locking when other write locks are present. Unlocking without locking should also fail."

        for f in lock_file_list:
            (tmp_path / f).touch()

        # Sanity check: lock files exist
        assert all((tmp_path / f).exists() for f in lock_file_list)

        for write in (True, False):
            self._ensure_lock_fails(tmp_path, write)

        self._ensure_unlock_fails(tmp_path)

    @pytest.mark.parametrize('lock_file_list',
                             [
                                 # One read lock
                                 ('read.ding.dong.lock',),
                                 # Two read locks
                                 ('read.dong.ding.lock', 'read.777.lock'),
                             ])
    def test_failing_read_lock(self, tmp_path, lock_file_list):
        "Test locking when other read locks are present. Unlocking without locking should also fail."

        for f in lock_file_list:
            (tmp_path / f).touch()

        # Sanity check: lock files exist
        assert all((tmp_path / f).exists() for f in lock_file_list)

        self._ensure_lock_fails(tmp_path, write=True)
        self._ensure_lock_unlock_succeeds(tmp_path, write=False)
        self._ensure_unlock_fails(tmp_path)

    @pytest.mark.parametrize('lock_file_list',
                             [
                                 # One write lock
                                 ('write.ding.dong.lock',),
                                 # Two write locks
                                 ('write.dong.ding.lock', 'write.777.lock'),
                                 # One write lock and one read lock
                                 ('write.ding.dong.lock', 'read.bing-bong.lock'),
                                 # 100 write and read locks
                                 tuple(f'write.{i}.lock' for i in range(100)) +
                                 tuple(f'read.{i}.lock' for i in range(100)),
                             ])
    def test_force_clear_all_locks(self, tmp_path, lock_file_list):
        "Test clearing all locks."

        for f in lock_file_list:
            (tmp_path / f).touch()

        # Sanity check: lock files exist
        assert all((tmp_path / f).exists() for f in lock_file_list)

        DirectoryLock(tmp_path).force_clear_all_locks()

        assert all(not (tmp_path / f).exists() for f in lock_file_list)



# Copyright 2021 Edward Leardi. All Rights Reserved.
#
# Copyright 2020 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import hashlib
from json import JSONDecodeError
import os
import pathlib

import pandas as pd
import pytest

from nourish.dataset import Dataset
from nourish.exceptions import DirectoryLockAcquisitionError
from nourish.loaders import FormatLoaderMap
from nourish.loaders.text import PlainTextLoader


class TestDataset:
    "Test Dataset class functionality."

    def test_mode(self, tmp_path, gmb_schema):
        "Test if Dataset class catches an invalid mode."

        with pytest.raises(ValueError) as e:
            Dataset(gmb_schema, data_dir=tmp_path, mode='DOWNLOAD_ONLY')
        assert str(e.value) == 'DOWNLOAD_ONLY not a valid mode'

    def test_data_dir(self, tmp_path, gmb_schema):
        "Test ``Dataset._data_dir``."
        # Automatic creation
        dataset = Dataset(gmb_schema, data_dir=tmp_path / 'data_dir', mode=Dataset.InitializationMode.LAZY)
        assert dataset._data_dir == tmp_path / 'data_dir'

        # Non-directory present
        dataset = Dataset(gmb_schema, data_dir='setup.py', mode=Dataset.InitializationMode.LAZY)
        with pytest.raises(NotADirectoryError) as e:
            dataset._data_dir
        assert str(e.value) == f'"{pathlib.Path.cwd()/"setup.py"}" exists and is not a directory.'

    def test_nourish_dir(self, tmp_path, gmb_schema):
        "Test ``Dataset._nourish_dir``."
        # Automatic creation
        nourish_dir = tmp_path / 'data_dir' / '.nourish.dataset'
        dataset = Dataset(gmb_schema, data_dir=tmp_path / 'data_dir', mode=Dataset.InitializationMode.LAZY)
        assert dataset._nourish_dir == nourish_dir
        # Non-directory present
        nourish_dir.rmdir()
        assert nourish_dir.exists() is False
        nourish_dir.touch()
        with pytest.raises(NotADirectoryError) as e:
            dataset._nourish_dir
        assert str(e.value) == f'"{nourish_dir}" exists and is not a directory.'

        # Non-directory parent present
        dataset = Dataset(gmb_schema, data_dir='setup.py', mode=Dataset.InitializationMode.LAZY)
        # These are raised by pathlib.Path.mkdir
        # Also see https://bugs.python.org/issue42872
        ExceptionClass = FileExistsError if os.name == 'nt' else NotADirectoryError
        with pytest.raises(ExceptionClass) as e:
            dataset._nourish_dir
        # This error message may be generated by pathlib.Path.mkdir() (as in DirectoryLock.lock()). We only make sure
        # the path is in the string.
        # On Windows, backslashes in the error message are doubled:
        # [WinError 183] Cannot create a file when that file already exists: 'D:\\\\a\\\\nourish\\\\nourish\\\\setup.py'
        assert str(pathlib.Path.cwd() / "setup.py").replace('\\', '\\\\') in str(e.value)

    def test_dataset_download(self, tmp_path, gmb_schema):
        "Test Dataset class downloads a dataset properly."

        data_dir = tmp_path / 'gmb'
        gmb_dataset = Dataset(gmb_schema, data_dir=data_dir, mode=Dataset.InitializationMode.DOWNLOAD_ONLY)
        assert len(list(data_dir.iterdir())) == 2  # 'groningen_meaning_bank_modified' and '.nourish.dataset'
        unarchived_data_dir = data_dir / 'groningen_meaning_bank_modified'
        unarchived_data_dir_files = ['gmb_subset_full.txt', 'LICENSE.txt', 'README.txt']
        assert unarchived_data_dir.is_dir()
        assert len(list(unarchived_data_dir.iterdir())) == len(unarchived_data_dir_files)
        assert all(f.name in unarchived_data_dir_files for f in unarchived_data_dir.iterdir())

        # Force check previously downloaded dataset should error
        with pytest.raises(RuntimeError) as e:
            gmb_dataset.download(check=True)
        assert str(e.value) == ('Dataset.download() was previously called. To overwrite existing data files, rerun '
                                'Dataset.download() with ``check`` set to ``False``.')

    def test_invalid_sha512(self, tmp_path, gmb_schema):
        "Test if Dataset class catches an invalid hash."

        gmb_schema['sha512sum'] = 'invalid hash example'

        with pytest.raises(IOError) as e:
            Dataset(gmb_schema, data_dir=tmp_path, mode=Dataset.InitializationMode.DOWNLOAD_ONLY)
        assert 'the file may by corrupted' in str(e.value)

    def test_load(self, downloaded_wikitext103_dataset):
        "Test basic loading functionality."

        data = downloaded_wikitext103_dataset.load()
        assert data is downloaded_wikitext103_dataset.data

        assert (hashlib.sha512(data['train'].encode()).hexdigest() ==
                ('df7615f77cb9dd19975881f271e3e3525bee38c08a67fea36a51c96be69a3ecabc9e05c02cbaf'
                 '6fc63a0082efb44156f61c81061d3b0272bbccd7657c682e791'))

        assert (hashlib.sha512(data['valid'].encode()).hexdigest() ==
                ('e4834d365d5f8313503895fd8304d29a566ff4a2df77efb32457fdc353304fb61460511f89bb9'
                 '0f14a47132c1539aaa324d3e71f5f56045a61a7292ad25a3c02'))

        assert (hashlib.sha512(data['test'].encode()).hexdigest() ==
                ('6fe665d33c0f788eba76da50539f0ca02432c70c94b788a493da491215e86043fc732dbeef9bb'
                 '49a72341c7283ea55f59d10941ac41f7ac58aea3bdcd72f5cd8'))

    def test_load_format_loader_map_param(self, downloaded_noaa_jfk_dataset):
        "Test ``format_loader_map`` param in ``Dataset.load``."

        # Default
        data = downloaded_noaa_jfk_dataset.load(format_loader_map=None)
        cleaned_data = data['jfk_weather_cleaned']
        assert isinstance(cleaned_data, pd.DataFrame)

        # Load CSV using the plain text loader
        data = downloaded_noaa_jfk_dataset.load(format_loader_map=FormatLoaderMap({
            'csv': PlainTextLoader()
        }))
        cleaned_data = data['jfk_weather_cleaned']
        assert isinstance(cleaned_data, str)

    def test_constructor_download_and_load(self, tmp_path, wikitext103_schema):
        "Test the full power of Dataset.__init__() (mode being ``InitializationMode.DOWNLOAD_AND_LOAD``)."

        dataset = Dataset(wikitext103_schema, data_dir=tmp_path, mode=Dataset.InitializationMode.DOWNLOAD_AND_LOAD)

        assert (hashlib.sha512(dataset.data['train'].encode()).hexdigest() ==
                ('df7615f77cb9dd19975881f271e3e3525bee38c08a67fea36a51c96be69a3ecabc9e05c02cbaf'
                 '6fc63a0082efb44156f61c81061d3b0272bbccd7657c682e791'))

        assert (hashlib.sha512(dataset.data['valid'].encode()).hexdigest() ==
                ('e4834d365d5f8313503895fd8304d29a566ff4a2df77efb32457fdc353304fb61460511f89bb9'
                 '0f14a47132c1539aaa324d3e71f5f56045a61a7292ad25a3c02'))

        assert (hashlib.sha512(dataset.data['test'].encode()).hexdigest() ==
                ('6fe665d33c0f788eba76da50539f0ca02432c70c94b788a493da491215e86043fc732dbeef9bb'
                 '49a72341c7283ea55f59d10941ac41f7ac58aea3bdcd72f5cd8'))

    def test_loading_undownloaded(self, tmp_path, gmb_schema):
        "Test loading before ``Dataset.download()`` has been called."

        dataset = Dataset(gmb_schema, data_dir=tmp_path, mode=Dataset.InitializationMode.LAZY)

        with pytest.raises(FileNotFoundError) as e:
            dataset.load(check=False)
        assert ('Failed to load subdataset "gmb_subset_full" because some files are not found. '
                'Did you forget to call Dataset.download()?\nCaused by:\n') in str(e.value)

        # Half-loaded data objects should get reset to None
        assert dataset._data is None
        with pytest.raises(RuntimeError) as e:
            dataset.data
        assert str(e.value) == ('Data has not been downloaded and/or loaded yet. Call Dataset.download() to download '
                                'data, call Dataset.load() to load data.')

        # Force check undownloaded dataset should error
        with pytest.raises(RuntimeError) as e:
            dataset.load(check=True)
        assert str(e.value) == (f'Downloaded data files are not present in {dataset._data_dir_} or are corrupted.')

    def test_unloaded_access_to_data(self, tmp_path, gmb_schema):
        "Test access to ``Dataset.data`` when no data has been loaded."

        dataset = Dataset(gmb_schema, data_dir=tmp_path, mode=Dataset.InitializationMode.LAZY)
        with pytest.raises(RuntimeError) as e:
            dataset.data
        assert str(e.value) == ('Data has not been downloaded and/or loaded yet. Call Dataset.download() to download '
                                'data, call Dataset.load() to load data.')

        # Same after downloading
        dataset.download()
        with pytest.raises(RuntimeError) as e:
            dataset.data
        assert str(e.value) == ('Data has not been downloaded and/or loaded yet. Call Dataset.download() to download '
                                'data, call Dataset.load() to load data.')

    def test_deleting_data_dir(self, tmp_path, gmb_schema):
        "Test ``Dataset.delete()``."

        # Note we don't use tmp_sub_dir fixture because we want data_dir to be non-existing at the beginning of the
        # test.
        data_dir = tmp_path / 'data-dir'
        dataset = Dataset(gmb_schema, data_dir=data_dir, mode=Dataset.InitializationMode.LAZY)
        assert not data_dir.exists()  # sanity check: data_dir doesn't exist
        dataset.delete()  # no exception should be raised here
        assert not data_dir.exists()  # sanity check: data_dir doesn't exist

        dataset.download()
        # Sanity check: Files are in place
        assert dataset.is_downloaded()
        assert len(os.listdir(data_dir)) > 0
        # Delete the dir
        dataset.delete()
        assert not data_dir.exists()

    def test_download_data_dir_is_not_a_dir(self, gmb_schema):
        "Test when downloading when ``data_dir`` exists and is not a dir."

        # These are raised by pathlib.Path.mkdir
        # Also see https://bugs.python.org/issue42872
        ExceptionClass = FileExistsError if os.name == 'nt' else NotADirectoryError
        with pytest.raises(ExceptionClass) as e:
            Dataset(gmb_schema, data_dir='./setup.py', mode=Dataset.InitializationMode.DOWNLOAD_ONLY)
        # This error message may be generated by pathlib.Path.mkdir() (as in DirectoryLock.lock()). We only make sure
        # the path is in the string.
        # On Windows, backslashes in the error message are doubled:
        # [WinError 183] Cannot create a file when that file already exists: 'D:\\\\a\\\\nourish\\\\nourish\\\\setup.py'
        assert str(pathlib.Path.cwd() / "setup.py").replace('\\', '\\\\') in str(e.value)

    def test_relative_data_dir(self, gmb_schema, chdir_tmp_path, tmp_sub_dir, tmp_relative_sub_dir):
        "Test when ``data_dir`` is relative."

        dataset = Dataset(gmb_schema, data_dir=tmp_relative_sub_dir, mode=Dataset.InitializationMode.LAZY)
        assert dataset._data_dir == tmp_sub_dir
        assert dataset._data_dir.is_absolute()

    def test_symlink_data_dir(self, tmp_symlink_dir, gmb_schema):
        "Test when ``data_dir`` is a symlink. The symlink should not be resolved."

        dataset = Dataset(gmb_schema, data_dir=tmp_symlink_dir, mode=Dataset.InitializationMode.LAZY)
        assert dataset._data_dir == tmp_symlink_dir

    def test_is_downloaded(self, tmp_path, gmb_schema):
        "Test is_downloaded method using a ``.tar.gz`` archive."

        data_dir = tmp_path / 'non-existing-dir'
        assert not data_dir.exists()  # Sanity check: data_dir must not exist
        gmb = Dataset(gmb_schema, data_dir=data_dir, mode=Dataset.InitializationMode.LAZY)
        assert gmb.is_downloaded() is False

        gmb.download()
        assert gmb.is_downloaded() is True

        # JSON decoding error
        gmb._file_list_file.write_text("nonsense\n", encoding='utf-8')
        with pytest.raises(JSONDecodeError):
            # We don't check the value of the exception because we clearly only are only interested in ensuring that the
            # file isn't decodable
            gmb.is_downloaded()

    def test_cache_dir_is_not_a_dir(self, tmp_path, gmb_schema):
        "Test when ``nourish_dir`` (i.e., ``data_dir/.nourish.dataset``) exists and is not a dir."
        (tmp_path / '.nourish.dataset').touch()  # Occupy this path with a regular file
        with pytest.raises(NotADirectoryError) as e:
            Dataset(gmb_schema, data_dir=tmp_path, mode=Dataset.InitializationMode.DOWNLOAD_ONLY)
        assert str(e.value) == f"\"{tmp_path/'.nourish.dataset'}\" exists and is not a directory."

    def _test_lock_exception(self, func, *, write, directory):
        "Helper for asserting lock exception."
        with pytest.raises(DirectoryLockAcquisitionError) as e:
            func()
        assert str(e.value) == f'Failed to acquire directory {"write" if write else "read"} lock for "{directory}"'

    def test_directory_write_lock_present(self, downloaded_gmb_dataset):
        "Test various functions when a write directory lock is present."

        lock_file = downloaded_gmb_dataset._nourish_dir_ / 'write.ding-dong.lock'
        lock_file.touch()

        f_stat = downloaded_gmb_dataset._file_list_file_.stat()
        self._test_lock_exception(
            lambda: downloaded_gmb_dataset.download(check=False),
            write=True,
            directory=downloaded_gmb_dataset._nourish_dir_)
        assert f_stat == downloaded_gmb_dataset._file_list_file_.stat()  # _file_list_file_ wasn't overwritten

        self._test_lock_exception(
            lambda: downloaded_gmb_dataset.load(), write=False, directory=downloaded_gmb_dataset._nourish_dir_)
        with pytest.raises(RuntimeError) as e:
            downloaded_gmb_dataset.data
        assert str(e.value) == ('Data has not been downloaded and/or loaded yet. Call Dataset.download() to download '
                                'data, call Dataset.load() to load data.')

        self._test_lock_exception(
            lambda: downloaded_gmb_dataset.delete(), write=True, directory=downloaded_gmb_dataset._nourish_dir_)
        assert downloaded_gmb_dataset._file_list_file_.exists()  # Files are still there
        downloaded_gmb_dataset.delete(force=True)
        # We purposefully use _file_list_file instead of _file_list_file_ to create the parent directory so that
        # TemporaryDirectory doesn't complain during test cleanup of downloaded_gmb_dataset (Python < 3.8)
        assert not downloaded_gmb_dataset._file_list_file.exists()

    def test_directory_read_lock_present(self, downloaded_gmb_dataset):
        "Test various functions when a directory read lock is present."

        lock_file = downloaded_gmb_dataset._nourish_dir_ / 'read.king-kong.lock'
        lock_file.touch()

        f_stat = downloaded_gmb_dataset._file_list_file_.stat()
        self._test_lock_exception(
            lambda: downloaded_gmb_dataset.download(check=False),
            write=True,
            directory=downloaded_gmb_dataset._nourish_dir_)
        assert f_stat == downloaded_gmb_dataset._file_list_file_.stat()  # _file_list_file_ wasn't overwritten

        downloaded_gmb_dataset.load()  # No exception raised
        assert downloaded_gmb_dataset.data is not None

        self._test_lock_exception(
            lambda: downloaded_gmb_dataset.delete(), write=True, directory=downloaded_gmb_dataset._nourish_dir_)
        assert downloaded_gmb_dataset._file_list_file_.exists()  # Files are still there
        downloaded_gmb_dataset.delete(force=True)
        # We purposefully use _file_list_file instead of _file_list_file_ to create the parent directory so that
        # TemporaryDirectory doesn't complain during test cleanup of downloaded_gmb_dataset (Python < 3.8)
        assert not downloaded_gmb_dataset._file_list_file.exists()



# Copyright 2021 Edward Leardi. All Rights Reserved.
#
# Copyright 2020 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest

from nourish import export_schemata, init, load_schemata
from nourish.schema import Schema
from nourish.exceptions import InsecureConnectionError
from nourish._schema_retrieval import retrieve_schema_file


class TestSchemaRetrieval:
    "Test schema retrieval."

    # We set tls_verification to False in this test. We don't test raising this warning because it is not within the
    # project's scope to guarantee that this warning will raise, and this warning raises only once, which means that
    # whether the warning will be raised depends on the order in which the tests are being executed.
    @pytest.mark.filterwarnings('ignore:Unverified HTTPS request .*:urllib3.exceptions.InsecureRequestWarning')
    @pytest.mark.parametrize('location_type',
                             ('absolute_dir',
                              'relative_dir',
                              'file_url',
                              'http_url',
                              'https_url'))
    def test_custom_schema_insecure(self, location_type, schema_file_relative_dir, request):
        "Test retrieving user-specified schema files, with insecure connections allowed."

        # We use '/' instead of os.path.sep because URLs only accept / not \ as separators, but Windows path accepts
        # both. This is not an issue for the purpose of this test.
        base = str(request.getfixturevalue('schema_file_' + location_type)) + '/'

        assert retrieve_schema_file(base + 'datasets.yaml', tls_verification=False) == \
            (schema_file_relative_dir / 'datasets.yaml').read_text(encoding='utf-8')
        assert retrieve_schema_file(base + 'formats.yaml', encoding='ascii', tls_verification=False) == \
            (schema_file_relative_dir / 'formats.yaml').read_text(encoding='ascii')
        with pytest.raises(UnicodeDecodeError) as e:
            retrieve_schema_file(base + 'licenses.yaml', encoding='ascii', tls_verification=False)
        # We usually don't assert the position info because we don't want to rewrite this test every time the yaml
        # file changes in length.
        assert "'ascii' codec can't decode byte 0xe2" in str(e.value)
        with pytest.raises(UnicodeDecodeError) as e:
            retrieve_schema_file(base + 'formats-utf-16le-bom.yaml', tls_verification=False)
        # Test "position 0" here because it should fail at the beginning of the decoding
        assert "'utf-8' codec can't decode byte 0xff in position 0" in str(e.value)
        with pytest.raises(UnicodeDecodeError) as e:
            retrieve_schema_file(base + 'formats-utf-16be.yaml', encoding='utf-8', tls_verification=False)
        assert "'utf-8' codec can't decode byte 0x90" in str(e.value)

    def test_invalid_schema(self):
        "Test retrieving user-specified invalid schema files."

        with pytest.raises(ValueError) as e:
            retrieve_schema_file(url_or_path='ftp://ftp/is/unsupported')

        assert str(e.value) == 'Unknown scheme in "ftp://ftp/is/unsupported": "ftp"'


class TestSecureSchemaRetrieval:
    "Common tests that mainly concern securely retrieving schema files."

    @pytest.fixture(params=('http_url', 'https_url'))
    def remote_dataset_schema_url(self, request) -> str:
        "http/https url to datasets.yaml."
        return str(request.getfixturevalue('schema_file_' + request.param)) + '/datasets.yaml'

    @pytest.fixture(params=('absolute_dir', 'relative_dir', 'file_url', 'https_url'))
    def dataset_schema_url_or_path(self, request) -> str:
        "Every kind of URL or path (other than http URL) to datasets.yaml."
        # We use '/' instead of os.path.sep because URLs only accept / not \ as separators, but Windows path accepts
        # both. This is not an issue for the purpose of this test.
        return str(request.getfixturevalue('schema_file_' + request.param)) + '/datasets.yaml'

    @pytest.mark.parametrize('caller', (retrieve_schema_file, Schema))
    def test_insecure_connections_fail(self, remote_dataset_schema_url, untrust_self_signed_cert, caller):
        "Test insecure connections that should fail."
        with pytest.raises(InsecureConnectionError) as e:
            caller(remote_dataset_schema_url, tls_verification=True)
        assert remote_dataset_schema_url in str(e.value)

    def test_secure_connections_succeed_retrieve_schema_file(self,
                                                             dataset_schema_url_or_path,
                                                             schema_file_relative_dir):
        "Test secure connections that should succeed for ``retrieve_schema_file``."
        assert retrieve_schema_file(dataset_schema_url_or_path, tls_verification=True) == \
            (schema_file_relative_dir / 'datasets.yaml').read_text(encoding='utf-8')

    def test_secure_connections_succeed_schema(self,
                                               dataset_schema_url_or_path,
                                               schema_file_relative_dir):
        "Test secure connections that should succeed for ``Schema()``."
        assert Schema(dataset_schema_url_or_path, tls_verification=True).retrieved_url_or_path == \
            dataset_schema_url_or_path

    def test_insecure_connections_load_schemata(self, remote_dataset_schema_url, untrust_self_signed_cert):
        "Test insecure connections that should fail when ``tls_verification=True`` for ``load_schemata``."
        init(update_only=True, DATASET_SCHEMA_URL=remote_dataset_schema_url)
        with pytest.raises(InsecureConnectionError) as e:
            load_schemata(force_reload=True, tls_verification=True)
        assert remote_dataset_schema_url in str(e.value)

        # Insecure load succeeds, no exception raised
        load_schemata(force_reload=True, tls_verification=False)
        assert export_schemata().schemata['datasets'].retrieved_url_or_path == remote_dataset_schema_url

    def test_secure_connections_succeed_load_schemata(self, dataset_schema_url_or_path):
        "Test secure connections that should succeed for :func:`nourish.load_schemata`."
        # We use '/' instead of os.path.sep because URLs only accept / not \ as separators, but Windows path accepts
        # both. This is not an issue for the purpose of this test.
        init(update_only=True, DATASET_SCHEMA_URL=dataset_schema_url_or_path)
        load_schemata(force_reload=True, tls_verification=True)
        assert export_schemata().schemata['datasets'].retrieved_url_or_path == dataset_schema_url_or_path

    @pytest.mark.parametrize('caller', (retrieve_schema_file, Schema.__init__, load_schemata))
    def test_default_tls_verification(self, caller):
        "Ensure that ``tls_verification=True`` by default."
        assert caller.__kwdefaults__['tls_verification'] is True



# Copyright 2021 Edward Leardi. All Rights Reserved.
#
# Copyright 2020 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pathlib
import setuptools


setuptools.setup(
    name="nourish",
    description="Consume and distribute datasets universally.",
    long_description=pathlib.Path('README.rst').read_text(),
    long_description_content_type="text/x-rst",
    keywords="machine-learning data-mining data-science",
    author="Edward Leardi",
    license="Apache v2",
    packages=setuptools.find_packages(),
    data_files=[("", ["LICENSE"])],
    python_requires=">=3.6",
    install_requires=[
        "dataclasses; python_version < '3.7.0'",  # backported dataclasses
        "packaging >= 20.4",
        "pandas >= 1.1.0",
        "pydantic >= 1.7.2",
        "PyYAML >= 5.3.1",
        "requests >= 2.24.0"],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development :: Libraries",
    ],
    use_scm_version={'write_to': 'nourish/_version.py'},
    setup_requires=['setuptools_scm']
)
